#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 07:23:52 (ywatanabe)"

from mngs.gen import deprecated


@deprecated("Use plot_with_ci(0 instaed.")
def fill_between(ax, xx, mean, std, label=None, alpha=0.1):
    ax.plot(xx, mean, label=label, alpha=alpha)
    ax.fill_between(xx, mean - std, mean + std, alpha=alpha)
    return ax


def plot_with_ci(ax, xx, mean, std, label=None, alpha=0.1):
    return fill_between(ax, xx, mean, std, label=None, alpha=0.1)



#!/usr/bin/env python3

from ._load import (
    load,
    load_configs,
    load_study_rdb,
    load_yaml_as_an_optuna_dict,
)
from ._reload import reload
from ._save import save, save_optuna_study_as_csv_and_pngs



# from ._yield_grid_params import yield_grid_params
from . import grid_search
from ._check_params import check_params
from ._DefaultDataset import DefaultDataset
from ._format_samples_for_sktime import format_samples_for_sktime
from ._LabelEncoder import LabelEncoder
from ._merge_labels import merge_labels
from ._sliding_window_data_augmentation import sliding_window_data_augmentation
from ._under_sample import under_sample
from ._verify_n_gpus import verify_n_gpus



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-05-30 09:35:17 (ywatanabe)"

"""
This script does XYZ.
"""

# Imports
import sys
from abc import abstractmethod

import matplotlib.pyplot as plt
import mngs
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from mngs.dsp.utils import build_bandpass_filters, init_bandpass_filters


class BaseFilter1D(nn.Module):
    def __init__(self, fp16=False, in_place=False):
        super().__init__()
        self.fp16 = fp16
        self.in_place = in_place
        # self.kernels = None

    @abstractmethod
    def init_kernels(
        self,
    ):
        """
        Abstract method to initialize filter kernels.
        Must be implemented by subclasses.
        """
        pass

    def forward(self, x, t=None, edge_len=0):
        """Apply the filter to input signal x with shape: (batch_size, n_chs, seq_len)"""

        # Shape check
        if self.fp16:
            x = x.half()

        x = mngs.dsp.ensure_3d(x)
        batch_size, n_chs, seq_len = x.shape

        # Kernel Check
        if self.kernels is None:
            raise ValueError("Filter kernels has not been initialized.")

        # Filtering
        x = self.flip_extend(x, self.kernel_size // 2)
        x = self.batch_conv(x, self.kernels, padding=0)
        x = x[..., :seq_len]

        assert x.shape == (
            batch_size,
            n_chs,
            len(self.kernels),
            seq_len,
        ), f"The shape of the filtered signal ({x.shape}) does not match the expected shape: ({batch_size}, {n_chs}, {len(self.kernels)}, {seq_len})."

        # Edge remove
        x = self.remove_edges(x, edge_len)

        if t is None:
            return x
        else:
            t = self.remove_edges(t, edge_len)
            return x, t

    @property
    def kernel_size(
        self,
    ):
        ks = self.kernels.shape[-1]
        # if not ks % 2 == 0:
        #     raise ValueError("Kernel size should be an even number.")
        return ks

    @staticmethod
    def flip_extend(x, extension_length):
        first_segment = x[:, :, :extension_length].flip(dims=[-1])
        last_segment = x[:, :, -extension_length:].flip(dims=[-1])
        return torch.cat([first_segment, x, last_segment], dim=-1)

    @staticmethod
    def batch_conv(x, kernels, padding="same"):
        """
        x: (batch_size, n_chs, seq_len)
        kernels: (n_kernels, seq_len_filt)
        """
        assert x.ndim == 3
        assert kernels.ndim == 2
        batch_size, n_chs, n_time = x.shape
        x = x.reshape(-1, x.shape[-1]).unsqueeze(1)
        kernels = kernels.unsqueeze(1)  # add the channel dimension
        n_kernels = len(kernels)
        filted = F.conv1d(x, kernels.type_as(x), padding=padding)
        return filted.reshape(batch_size, n_chs, n_kernels, -1)

    @staticmethod
    def remove_edges(x, edge_len):
        edge_len = x.shape[-1] // 8 if edge_len == "auto" else edge_len

        if 0 < edge_len:
            return x[..., edge_len:-edge_len]
        else:
            return x


class BandPassFilter(BaseFilter1D):
    def __init__(self, bands, fs, seq_len, fp16=False):
        super().__init__(fp16=fp16)

        self.fp16 = fp16

        # Ensures bands shape
        assert bands.ndim == 2

        # Check bands definitions
        nyq = fs / 2.0
        bands = torch.clip(bands, 0.1, nyq - 1)
        for ll, hh in bands:
            assert 0 < ll
            assert ll < hh
            assert hh < nyq

        # Prepare kernels
        kernels = self.init_kernels(seq_len, fs, bands)
        if fp16:
            kernels = kernels.half()
        self.register_buffer(
            "kernels",
            kernels,
        )

    @staticmethod
    def init_kernels(seq_len, fs, bands):
        filters = [
            mngs.dsp.utils.design_filter(
                seq_len,
                fs,
                low_hz=ll,
                high_hz=hh,
                is_bandstop=False,
            )
            for ll, hh in bands
        ]

        kernels = mngs.dsp.utils.zero_pad(filters)
        kernels = mngs.dsp.utils.ensure_even_len(kernels)
        # kernels = torch.tensor(kernels).clone().detach()
        kernels = kernels.clone().detach().requires_grad_(True)
        return kernels


# /home/ywatanabe/proj/mngs/src/mngs/nn/_Filters.py:155: UserWarning: To copy construct from a tensor, it is recommended to use sourceTensor.clone().detach() or sourceTensor.clone().detach().requires_grad_(True), rather than torch.tensor(sourceTensor).
#   kernels = torch.tensor(kernels).clone().detach()


class BandStopFilter(BaseFilter1D):
    def __init__(self, bands, fs, seq_len):
        super().__init__()

        # Ensures bands shape
        assert bands.ndim == 2

        # Check bands definitions
        nyq = fs / 2.0
        bands = np.clip(bands, 0.1, nyq - 1)
        for ll, hh in bands:
            assert 0 < ll
            assert ll < hh
            assert hh < nyq

        self.register_buffer("kernels", self.init_kernels(seq_len, fs, bands))

    @staticmethod
    def init_kernels(seq_len, fs, bands):
        kernels = mngs.dsp.utils.zero_pad(
            [
                mngs.dsp.utils.design_filter(
                    seq_len, fs, low_hz=ll, high_hz=hh, is_bandstop=True
                )
                for ll, hh in bands
            ]
        )
        kernels = mngs.dsp.utils.ensure_even_len(kernels)
        return torch.tensor(kernels)


class LowPassFilter(BaseFilter1D):
    def __init__(self, cutoffs_hz, fs, seq_len):
        super().__init__()

        # Ensures bands shape
        assert cutoffs_hz.ndim == 1

        # Check bands definitions
        nyq = fs / 2.0
        bands = np.clip(cutoffs_hz, 0.1, nyq - 1)
        for cc in cutoffs_hz:
            assert 0 < cc
            assert cc < nyq

        self.register_buffer(
            "kernels", self.init_kernels(seq_len, fs, cutoffs_hz)
        )

    @staticmethod
    def init_kernels(seq_len, fs, cutoffs_hz):
        kernels = mngs.dsp.utils.zero_pad(
            [
                mngs.dsp.utils.design_filter(
                    seq_len, fs, low_hz=None, high_hz=cc, is_bandstop=False
                )
                for cc in cutoffs_hz
            ]
        )
        kernels = mngs.dsp.utils.ensure_even_len(kernels)
        return torch.tensor(kernels)


class HighPassFilter(BaseFilter1D):
    def __init__(self, cutoffs_hz, fs, seq_len):
        super().__init__()

        # Ensures bands shape
        assert cutoffs_hz.ndim == 1

        # Check bands definitions
        nyq = fs / 2.0
        bands = np.clip(cutoffs_hz, 0.1, nyq - 1)
        for cc in cutoffs_hz:
            assert 0 < cc
            assert cc < nyq

        self.register_buffer(
            "kernels", self.init_kernels(seq_len, fs, cutoffs_hz)
        )

    @staticmethod
    def init_kernels(seq_len, fs, cutoffs_hz):
        kernels = mngs.dsp.utils.zero_pad(
            [
                mngs.dsp.utils.design_filter(
                    seq_len, fs, low_hz=cc, high_hz=None, is_bandstop=False
                )
                for cc in cutoffs_hz
            ]
        )
        kernels = mngs.dsp.utils.ensure_even_len(kernels)
        return torch.tensor(kernels)


class GaussianFilter(BaseFilter1D):
    def __init__(self, sigma):
        super().__init__()
        self.sigma = mngs.gen.to_even(sigma)
        self.register_buffer("kernels", self.init_kernels(sigma))

    @staticmethod
    def init_kernels(sigma):
        kernel_size = sigma * 6  # +/- 3SD
        kernel_range = torch.arange(0, kernel_size) - kernel_size // 2
        kernel = torch.exp(-0.5 * (kernel_range / sigma) ** 2)
        kernel /= kernel.sum()
        kernels = kernel.unsqueeze(0)  # n_filters = 1
        kernels = mngs.dsp.utils.ensure_even_len(kernels)
        return torch.tensor(kernels)


class DifferentiableBandPassFilter(BaseFilter1D):
    def __init__(
        self,
        sig_len,
        fs,
        pha_low_hz=2,
        pha_high_hz=20,
        pha_n_bands=30,
        amp_low_hz=80,
        amp_high_hz=160,
        amp_n_bands=50,
        cycle=3,
        fp16=False,
    ):
        super().__init__(fp16=fp16)

        # Attributes
        self.pha_low_hz = pha_low_hz
        self.pha_high_hz = pha_high_hz
        self.amp_low_hz = amp_low_hz
        self.amp_high_hz = amp_high_hz
        self.sig_len = sig_len
        self.fs = fs
        self.cycle = cycle
        self.fp16 = fp16

        # Check bands definitions
        nyq = fs / 2.0
        pha_high_hz = torch.tensor(pha_high_hz).clip(0.1, nyq - 1)
        pha_low_hz = torch.tensor(pha_low_hz).clip(0.1, pha_high_hz - 1)
        amp_high_hz = torch.tensor(amp_high_hz).clip(0.1, nyq - 1)
        amp_low_hz = torch.tensor(amp_low_hz).clip(0.1, amp_high_hz - 1)

        assert pha_low_hz < pha_high_hz < nyq
        assert amp_low_hz < amp_high_hz < nyq

        # Prepare kernels
        self.init_kernels = init_bandpass_filters
        self.build_bandpass_filters = build_bandpass_filters
        kernels, self.pha_mids, self.amp_mids = self.init_kernels(
            sig_len=sig_len,
            fs=fs,
            pha_low_hz=pha_low_hz,
            pha_high_hz=pha_high_hz,
            pha_n_bands=pha_n_bands,
            amp_low_hz=amp_low_hz,
            amp_high_hz=amp_high_hz,
            amp_n_bands=amp_n_bands,
            cycle=cycle,
        )

        self.register_buffer(
            "kernels",
            kernels,
        )
        # self.register_buffer("pha_mids", pha_mids)
        # self.register_buffer("amp_mids", amp_mids)
        # self.pha_mids = nn.Parameter(pha_mids.detach())
        # self.amp_mids = nn.Parameter(amp_mids.detach())

        if fp16:
            self.kernels = self.kernels.half()
            # self.pha_mids = self.pha_mids.half()
            # self.amp_mids = self.amp_mids.half()

    def forward(self, x, t=None, edge_len=0):
        # Constrains the parameter spaces
        torch.clip(self.pha_mids, self.pha_low_hz, self.pha_high_hz)
        torch.clip(self.amp_mids, self.amp_low_hz, self.amp_high_hz)

        self.kernels = self.build_bandpass_filters(
            self.sig_len, self.fs, self.pha_mids, self.amp_mids, self.cycle
        )
        return super().forward(x=x, t=t, edge_len=edge_len)


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, fig_scale=5
    )

    xx, tt, fs = mngs.dsp.demo_sig(sig_type="chirp")
    xx = torch.tensor(xx).cuda()
    # bands = np.array([[2, 3], [3, 4]])
    # BandPassFilter(bands, fs, xx.shape)
    m = DifferentiableBandPassFilter(xx.shape[-1], fs).cuda()

    mngs.ml.utils.check_params(m)
    # {'pha_mids': (torch.Size([30]), 'Learnable'),
    #  'amp_mids': (torch.Size([50]), 'Learnable')}

    xf = m(xx)  # (8, 19, 80, 2048)

    xf.sum().backward()  # OK, differentiable

    m.pha_mids
    # Parameter containing:
    # tensor([ 2.0000,  2.6207,  3.2414,  3.8621,  4.4828,  5.1034,  5.7241,  6.3448,
    #          6.9655,  7.5862,  8.2069,  8.8276,  9.4483, 10.0690, 10.6897, 11.3103,
    #         11.9310, 12.5517, 13.1724, 13.7931, 14.4138, 15.0345, 15.6552, 16.2759,
    #         16.8966, 17.5172, 18.1379, 18.7586, 19.3793, 20.0000],
    #        requires_grad=True)
    m.amp_mids
    # Parameter containing:
    # tensor([ 80.0000,  81.6327,  83.2653,  84.8980,  86.5306,  88.1633,  89.7959,
    #          91.4286,  93.0612,  94.6939,  96.3265,  97.9592,  99.5918, 101.2245,
    #         102.8571, 104.4898, 106.1225, 107.7551, 109.3878, 111.0204, 112.6531,
    #         114.2857, 115.9184, 117.5510, 119.1837, 120.8163, 122.4490, 124.0816,
    #         125.7143, 127.3469, 128.9796, 130.6122, 132.2449, 133.8775, 135.5102,
    #         137.1429, 138.7755, 140.4082, 142.0408, 143.6735, 145.3061, 146.9388,
    #         148.5714, 150.2041, 151.8367, 153.4694, 155.1020, 156.7347, 158.3673,
    #         160.0000], requires_grad=True)

    # PSD
    bands = torch.hstack([m.pha_mids, m.amp_mids])

    # Plots PSD
    # matplotlib.use("TkAgg")
    fig, axes = mngs.plt.subplots(nrows=1 + len(bands), ncols=2)

    psd, ff = mngs.dsp.psd(xx, fs)  # Orig
    axes[0, 0].plot(tt, xx[0, 0].detach().cpu().numpy(), label="orig")
    axes[0, 1].plot(
        ff.detach().cpu().numpy(),
        psd[0, 0].detach().cpu().numpy(),
        label="orig",
    )

    for i_filt in range(len(bands)):
        mid_hz = int(bands[i_filt].item())
        psd_f, ff_f = mngs.dsp.psd(xf[:, :, i_filt, :], fs)
        axes[i_filt + 1, 0].plot(
            tt,
            xf[0, 0, i_filt].detach().cpu().numpy(),
            label=f"filted at {mid_hz} Hz",
        )
        axes[i_filt + 1, 1].plot(
            ff_f.detach().cpu().numpy(),
            psd_f[0, 0].detach().cpu().numpy(),
            label=f"filted at {mid_hz} Hz",
        )
    for ax in axes.ravel():
        ax.legend(loc="upper left")

    mngs.io.save(fig, "traces.png")
    # plt.show()

    # Close
    mngs.gen.close(CONFIG)

"""
/home/ywatanabe/proj/entrance/mngs/dsp/nn/_Filters.py
"""



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-21 12:05:35"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

import os
import sys

import matplotlib.pyplot as plt

# Imports
import mngs
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F

# # Config
# CONFIG = mngs.gen.load_configs()

# Functions
def less(output):
    """
    Print the given output using `less` in an IPython or IPdb session.
    """
    import os
    import tempfile

    from IPython import get_ipython

    # Create a temporary file to hold the output
    with tempfile.NamedTemporaryFile(delete=False, mode="w+t") as tmpfile:
        # Write the output to the temporary file
        tmpfile.write(output)
        tmpfile_name = tmpfile.name

    # Use IPython's system command access to pipe the content of the temporary file to `less`
    get_ipython().system(f"less {tmpfile_name}")

    # Clean up the temporary file
    os.remove(tmpfile_name)


# (YOUR AWESOME CODE)

if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/gen/_less.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-02 21:28:58 (ywatanabe)"

"""
This script does XYZ.
"""

import sys

import matplotlib.pyplot as plt
import mngs
import torch
from mngs.general import torch_fn
from mngs.nn import Hilbert


# Functions
@torch_fn
def hilbert(
    x,
    dim=-1,
):
    y = Hilbert(x.shape[-1], dim=dim)(x)
    return y[..., 0], y[..., 1]


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # Parameters
    T_SEC = 1.0
    FS = 400
    SIG_TYPE = "chirp"

    # Demo signal
    xx, tt, fs = mngs.dsp.demo_sig(t_sec=T_SEC, fs=FS, sig_type=SIG_TYPE)

    # Main
    pha, amp = hilbert(
        xx,
        dim=-1,
    )
    # (32, 19, 1280, 2)

    # Plots
    fig, axes = mngs.plt.subplots(nrows=2, sharex=True)
    fig.suptitle("Hilbert Transformation")

    axes[0].plot(tt, xx[0, 0], label=SIG_TYPE)
    axes[0].plot(tt, amp[0, 0], label="Amplidue")
    axes[0].legend()
    # axes[0].set_xlabel("Time [s]")
    axes[0].set_ylabel("Amplitude [?V]")

    axes[1].plot(tt, pha[0, 0], label="Phase")
    axes[1].legend()

    axes[1].set_xlabel("Time [s]")
    axes[1].set_ylabel("Phase [rad]")

    # plt.show()
    mngs.io.save(fig, "traces.png")

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/dsp/_hilbert.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-30 18:17:46 (ywatanabe)"

import mngs
import numpy as np
import torch


def mean(x, axis=-1, dim=None):
    if dim is not None:
        axis = dim
    return x.mean(axis, keepdims=True)


def std(x, axis=-1, dim=None):
    if dim is not None:
        axis = dim
    return x.std(axis, keepdims=True)


def zscore(x, axis=-1, dim=None):
    if dim is not None:
        axis = dim

    x, x_type = mngs.gen.my2tensor(x)

    _mean = mean(x, axis=axis)
    diffs = x - _mean
    var = torch.mean(torch.pow(diffs, 2.0), dim=axis, keepdims=True)
    std = torch.pow(var, 0.5)
    out = diffs / std

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def kurtosis(x, axis=-1, dim=None):
    if dim is not None:
        axis = dim

    x, x_type = mngs.gen.my2tensor(x)

    zscores = zscore(x, axis=axis)
    out = torch.mean(torch.pow(zscores, 4.0), dim=axis, keepdims=True) - 3.0

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def skewness(x, axis=-1, dim=None):
    if dim is not None:
        axis = dim

    x, x_type = mngs.gen.my2tensor(x)

    zscores = zscore(x, dim=axis)
    out = torch.mean(torch.pow(zscores, 3.0), dim=axis, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def median(x, axis=-1, dim=None):
    if dim is not None:
        axis = dim

    x, x_type = mngs.gen.my2tensor(x)

    out = torch.median(x, dim=axis, keepdims=True)[0]

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def q(x, q, axis=-1, dim=None):
    if dim is not None:
        axis = dim

    x, x_type = mngs.gen.my2tensor(x)

    out = torch.quantile(x, q / 100, dim=axis, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


if __name__ == "__main__":
    x = np.random.rand(4, 3, 2)
    mean(x)
    std(x)
    zscore(x)
    kurtosis(x)
    skewness(x)
    median(x)
    q(x, 25)

    x = torch.tensor(x)
    mean(x)
    std(x)
    zscore(x)
    kurtosis(x)
    skewness(x)
    median(x)
    q(x, 25)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-05-04 16:54:55 (ywatanabe)"

#!/usr/bin/env python

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchsummary import summary
import mngs

MNet_config = {
    "classes": ["class1", "class2"],
    "n_chs": 270,
    "n_fc1": 1024,
    "d_ratio1": 0.85,
    "n_fc2": 256,
    "d_ratio2": 0.85,
}


class MNet_1000(nn.Module):
    def __init__(self, config):
        super().__init__()

        # basic
        self.config = config
        # fc
        N_FC_IN = 15950


        # conv
        self.backborn = nn.Sequential(*[
            
            nn.Conv2d(1, 40, kernel_size=(config["n_chs"], 4)),
            nn.Mish(),

            nn.Conv2d(40, 40, kernel_size=(1, 4)),
            nn.BatchNorm2d(40),
            nn.MaxPool2d((1, 5)),
            nn.Mish(),

            SwapLayer(),

            nn.Conv2d(1, 50, kernel_size=(8, 12)),
            nn.BatchNorm2d(50),
            nn.MaxPool2d((3, 3)),
            nn.Mish(),

            nn.Conv2d(50, 50, kernel_size=(1, 5)),
            nn.BatchNorm2d(50),
            nn.MaxPool2d((1, 2)),
            nn.Mish(),

            ReshapeLayer(),
            nn.Linear(N_FC_IN, config["n_fc1"]),
            ])
        
        # # conv
        # self.conv1 = nn.Conv2d(1, 40, kernel_size=(config["n_chs"], 4))
        # self.act1 = nn.Mish()

        # self.conv2 = nn.Conv2d(40, 40, kernel_size=(1, 4))
        # self.bn2 = nn.BatchNorm2d(40)
        # self.pool2 = nn.MaxPool2d((1, 5))
        # self.act2 = nn.Mish()

        # self.swap = SwapLayer()

        # self.conv3 = nn.Conv2d(1, 50, kernel_size=(8, 12))
        # self.bn3 = nn.BatchNorm2d(50)
        # self.pool3 = nn.MaxPool2d((3, 3))
        # self.act3 = nn.Mish()

        # self.conv4 = nn.Conv2d(50, 50, kernel_size=(1, 5))
        # self.bn4 = nn.BatchNorm2d(50)
        # self.pool4 = nn.MaxPool2d((1, 2))
        # self.act4 = nn.Mish()


        self.fc = nn.Sequential(
            # nn.Linear(N_FC_IN, config["n_fc1"]),
            nn.Mish(),
            nn.Dropout(config["d_ratio1"]),
            nn.Linear(config["n_fc1"], config["n_fc2"]),
            nn.Mish(),
            nn.Dropout(config["d_ratio2"]),
            nn.Linear(config["n_fc2"], len(config["classes"])),
        )

    @staticmethod
    def _reshape_input(x, n_chs):
        """
        (batch, channel, time_length) -> (batch, channel, time_length, new_axis)
        """
        if x.ndim == 3:
            x = x.unsqueeze(-1)
        if x.shape[2] == n_chs:
            x = x.transpose(1, 2)
        x = x.transpose(1, 3).transpose(2, 3)
        return x

    @staticmethod
    def _znorm_along_the_last_dim(x):
        return (x - x.mean(dim=-1, keepdims=True)) / x.std(dim=-1, keepdims=True)

    def forward(self, x):
        # # time-wise normalization
        # x = self._znorm_along_the_last_dim(x)
        # x = self._reshape_input(x, self.config["n_chs"])

        # x = self.backborn(x)
        x = self.forward_bb(x)

        # x = x.reshape(len(x), -1)

        x = self.fc(x)

        return x

    def forward_bb(self, x):
        # time-wise normalization
        x = self._znorm_along_the_last_dim(x)
        x = self._reshape_input(x, self.config["n_chs"])
        x = self.backborn(x)
        return x
        

    

class SwapLayer(nn.Module):
    def __init__(
        self,
    ):
        super().__init__()

    def forward(self, x):
        return x.transpose(1, 2)
    
class ReshapeLayer(nn.Module):
    def __init__(
        self,
    ):
        super().__init__()

    def forward(self, x):
        return x.reshape(len(x), -1)


if __name__ == "__main__":
    ## Demo data
    BS, N_CHS, SEQ_LEN = 16, 270, 1000
    x = torch.rand(BS, N_CHS, SEQ_LEN).cuda()

    ## Config for the model
    model = MNet_1000(MNet_config).cuda()

    y = model(x)
    summary(model, x)
    print(y.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2021-10-26 21:17:18 (ylab)"

import re
from codecs import open
from os import path

from setuptools import find_packages, setup

################################################################################
PACKAGE_NAME = "mngs"
# PACKAGE_DATA = {'data': ['']}
# DATA_FILES = [('my_data, [data/data_file']),]
PACKAGES = find_packages(where="src")
DESCRIPTION = (
    "For lazy python users (monogusa people in Japanse), especially in ML/DSP fields"
)
KEYWORDS = ["utils", "utilities", "python", "machine learning"]
CLASSIFIERS = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
]
################################################################################

root_dir = path.abspath(path.dirname(__file__))


def _requirements():
    return [
        name.rstrip()
        for name in open(path.join(root_dir, "requirements.txt")).readlines()
    ]


def _test_requirements():
    return [
        name.rstrip()
        for name in open(path.join(root_dir, "test-requirements.txt")).readlines()
    ]


with open(path.join(root_dir, "src", PACKAGE_NAME, "__init__.py")) as f:
    init_text = f.read()
    version = re.search(r"__version__\s*=\s*[\'\"](.+?)[\'\"]", init_text).group(1)
    license = re.search(r"__license__\s*=\s*[\'\"](.+?)[\'\"]", init_text).group(1)
    author = re.search(r"__author__\s*=\s*[\'\"](.+?)[\'\"]", init_text).group(1)
    author_email = re.search(
        r"__author_email__\s*=\s*[\'\"](.+?)[\'\"]", init_text
    ).group(1)
    url = re.search(r"__url__\s*=\s*[\'\"](.+?)[\'\"]", init_text).group(1)

assert version
assert license
assert author
assert author_email
assert url

# with open("README.rst", encoding="utf-8") as f:
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    package_dir={"": "src"},
    name=PACKAGE_NAME,
    packages=PACKAGES,
    # py_modules=PY_MODULES,
    # package_data=PACKAGE_DATA,
    # data_files=DATA_FILES,
    version=version,
    license=license,
    install_requires=_requirements(),
    tests_require=_test_requirements(),
    setup_requires=["pytest-runner"],
    author=author,
    author_email=author_email,
    url=url,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    python_requires=">=3.0",
)



#!/usr/bin/env python3

# from . import colors
from ._PARAMS import PARAMS

_ = None
from . import ax
from ._add_hue import add_hue
from ._annotated_heatmap import annotated_heatmap
from ._colors import (
    cycle_color,
    rgb2rgba,
    rgba2hex,
    to_hex,
    to_rgb,
    to_rgba,
    update_alpha,
)
from ._configure_mpl import configure_mpl
from ._draw_a_cube import draw_a_cube
from ._get_RGBA_from_colormap import get_RGBA_from_colormap
from ._grid_image import grid_image
from ._interp_colors import interp_colors
from ._mk_colorbar import mk_colorbar
from ._mk_patches import mk_patches
from ._subplots._SubplotsManager import subplots
from ._tpl import termplot



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import random
import sys
import warnings
from copy import deepcopy
from functools import partial
from pprint import pprint
from time import sleep

import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd
import sklearn
import torch
from natsort import natsorted
try:
    from pandas.core.common import SettingWithCopyWarning # 1.4
except:
    from pandas.errors import SettingWithCopyWarning # 1.5
from sklearn.preprocessing import power_transform
from torch.utils.data import DataLoader, Dataset

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

sys.path.append("./externals")
from dataloader.load_BIDS import load_BIDS as load_BIDS_func
from dataloader.utils.divide_into_datasets import divide_into_datasets
from dataloader.utils.label_handlers import _expand_conc_classes_str


class DataLoaderFiller(object):
    """
    Fills cropped EEG samples and their labels into self.dl_tra, self.dl_val, and self.dl_tes,
    each of which is torch.utils.data.DataLoader. These data loaders yields batches,
    and every batch contains the followings in this order:

        Xb: batched inputs, EEG (n_samples * n_channel * seq_len)
        Tb: batched targets, Diagnosed labels
        Sgb: batched Subjects ID (global)
        Slb: batched Subjects ID, reallocated within a dataset like Training Dataset (local)

    Examples:

    # ----------------------------------------
    # NN Training
    # ----------------------------------------
    BIDS_ROOT = (
        f"/storage/dataset/EEG/internal/EEG_DiagnosisFromRawSignals/"
        f"BIDS_dataset_v5.2_Osk"
    )
    disease_types=["HV", "AD", "NPH", "DLB"]

    dlf = DataLoaderFiller(BIDS_ROOT, disease_types)

    n_epochs = 3
    for i_fold in range(dlf.filler_params["num_folds"]):
        dlf.fill(i_fold, reset_fill_counter=True)
        for epoch in range(n_epochs):
            dlf.fill(i_fold, reset_fill_counter=False)

            for i_batch, batch in enumerate(dlf.dl_val):
                Xb, Tb, Sgb, Slb = batch
                # Validate your model

            for i_batch, batch in enumerate(dlf.dl_tra):
                Xb, Tb, Sgb, Slb = batch
                # Train your model

        for i_batch, batch in enumerate(dlf.dl_tes):
            Xb, Tb, Sgb, Slb = batch
            # Test your model

    # Summarize Cross Validation Scores

    # ----------------------------------------
    # Makes a dataloader from an existing BIDS layout
    # ----------------------------------------
    BIDS_ROOT = (
        f"/storage/dataset/EEG/internal/EEG_DiagnosisFromRawSignals/"
        f"BIDS_dataset_v5.2_Osk"
    )
    load_params, data_all, data_not_used = DataLoaderFiller.load_BIDS(
        BIDS_ROOT,
        disease_types=["HV", "AD", "NPH", "DLB"],
        from_pkl=True,
        samp_rate=500,
    )
    subs_labels_uq = data_all[["subject", "disease_type"]].drop_duplicates
    window_size_pts = 1000
    sub_str2int_local = {s:i for i,s in enumerate(subs_labels_uq["subject"])}
    sub_str2int_global = sub_str2int_local

    dl = DataLoaderFiller.crop_without_overlap(
        data_all,
        subs_labels_uq["subject"],
        subs_labels_uq["disease_type"],
        window_size_pts,
        sub_str2int_val,
        sub_str2int_global,
        crop_num=None,
        dtype_np=np.float32,
        transform=None,
        field_order=None,
        batch_size=32,
        num_workers=16,
    )

    # ----------------------------------------
    # Directly accesses the dataset
    # ----------------------------------------
    X_tra, T_tra, Sg_tra, Sl_tra = dlf.dl_tra.dataset.arrs_list
    X_val, T_val, Sg_val, Sl_val = dlf.dl_val.dataset.arrs_list
    X_tes, T_tes, Sg_tes, Sl_tes = dlf.dl_tes.dataset.arrs_list

    Note:
        dir(dlf) or dlf.[TAB KEY] will show you availabel attributes and methods.
        Especially,

    """

    def __init__(
        self,
        BIDS_ROOT,
        disease_types,
        stop_bands=None,
        val_ratio_to_trn=None,
        is_no_test=None,
        drop_cMCI=False,
        window_size_pts=1000,
        montage=None,
    ):

        self.load_params, self.data_all, self.data_not_used = self.load_BIDS(
            BIDS_ROOT,
            disease_types=disease_types,
            samp_rate=500,
            drop_cMCI=drop_cMCI,
            montage=montage,
        )

        # self.data_all = self._process_with_demographic_data(self.data_all)

        if val_ratio_to_trn is not None:
            self.load_params["val_ratio_to_trn"] = val_ratio_to_trn  # override

        if is_no_test is True:
            self.load_params["num_folds"] = 1  # override

        if stop_bands is not None:
            for band_str in stop_bands:
                self.data_all = self._band_stop(self.data_all, band_str)

        self.filler_params = self._load_and_process_filler_params(
            self.load_params["conc_classes"]
        )
        self.filler_params["window_size_pts"] = window_size_pts
        self.cv_dict = self._mk_cv_dict()

        self.db = self._mk_db(self.data_all, self.cv_dict)
        self._fix_seeds(random_state=self.filler_params["random_state"])
        self._fill_counter = 0

    def _band_stop(self, data_all, band_str):
        def _inner_band_stop(low_hz, high_hz, eeg_record):
            return np.array(
                mngs.dsp.bandstop(
                    torch.tensor(eeg_record).unsqueeze(0).cuda(),
                    samp_rate=500,
                    low_hz=low_hz,
                    high_hz=high_hz,
                ).cpu()
            ).squeeze()

        BANDS_LIM_HZ = mngs.io.load("./config/global.yaml")["BANDS_LIM_HZ_DICT"]
        low_hz, high_hz = BANDS_LIM_HZ[band_str]

        p_inner_band_stop = partial(_inner_band_stop, low_hz, high_hz)

        print(f"\n{band_str} band is being suppressed ...\n")
        self.data_all["eeg"] = self.data_all["eeg"].apply(p_inner_band_stop)

        return self.data_all

    def _process_data_all(self, data_all):
        load_params["conc_classes"] = disease_types  # e.g., ["AD", "DLB", "HV", "NPH"]
        expanded_classes = np.hstack(_expand_conc_classes_str(disease_types))

        indi_to_use = (
            pd.concat(
                [data_all["disease_type"] == ec for ec in expanded_classes], axis=1
            )
            .sum(axis=1)
            .astype(bool)
        )

        data_not_used = data_all.loc[~indi_to_use]
        data_all = data_all.loc[indi_to_use]

    def _mk_cv_dict(
        self,
    ):
        div_dict = {}
        keys = [
            "class_field_name",
            "num_folds",
            "conc_classes",
            "val_ratio_to_trn",
            "undersample_trn",
            "undersample_val",
            "random_state",
        ]

        for k in keys:
            try:
                div_dict[k] = self.filler_params[k]
            except:
                div_dict[k] = self.load_params[k]

        cv_dict = divide_into_datasets(self.data_all, div_dict)

        return cv_dict

    def fill(self, i_fold, reset_fill_counter=True):
        """
        Fills self.dl_tra, self.dl_val, and self.dl_tes.

        self.dl_val and self.dl_tes are always the identical ones because they have no randomness.
        """
        self.i_fold = i_fold

        if reset_fill_counter:
            self._fill_counter = 0
            self._check_cv_health(self.cv_dict, i_fold)

        self.dl_tra = self._fill_twice_length_trn_ds(
            self.data_all,
            self.cv_dict["label_int_2_conc_class_dict"],
            self.subs_trn,
            self.labels_trn,
            self.sub_str2int_trn,
            self.sub_str2int_global,
            self.filler_params["window_size_pts"],
            dtype_np=self.filler_params["_dtype_np"],
            field_order=self.filler_params["field_order"],
            batch_size=self.filler_params["batch_size"],
            num_workers=self.filler_params["num_workers"],
        )

        if self._fill_counter == 0:

            if self.load_params["val_ratio_to_trn"] == 0:
                self._dl_val = None
            else:
                self._dl_val = self.crop_without_overlap(
                    self.data_all,
                    self.subs_val,
                    self.labels_val,
                    self.filler_params["window_size_pts"],
                    self.sub_str2int_val,
                    self.sub_str2int_global,
                    crop_num=None,
                    dtype_np=self.filler_params["_dtype_np"],
                    transform=self.filler_params["transform"],
                    field_order=self.filler_params["field_order"],
                    batch_size=self.filler_params["batch_size"],
                    num_workers=self.filler_params["num_workers"],
                )

            self._dl_tes = self.crop_without_overlap(
                self.data_all,
                self.subs_tes,
                self.labels_tes,
                self.filler_params["window_size_pts"],
                self.sub_str2int_tes,
                self.sub_str2int_global,
                crop_num=None,
                dtype_np=self.filler_params["_dtype_np"],
                transform=self.filler_params["transform"],
                field_order=self.filler_params["field_order"],
                batch_size=self.filler_params["batch_size"],
                num_workers=self.filler_params["num_workers"],
            )

            self.dl_val = deepcopy(self._dl_val)
            self.dl_tes = deepcopy(self._dl_tes)

        if self._fill_counter >= 1:
            self.dl_val = deepcopy(self._dl_val)
            self.dl_tes = deepcopy(self._dl_tes)

        self._fill_counter += 1

    def _fill_twice_length_trn_ds(
        self,
        data_all,
        label2class_dict,
        subs_trn,
        labels_trn,
        subj_str2int_trn,
        subj_str2int_global,
        window_size_pts,
        dtype_np=np.float32,
        transform=None,
        field_order=None,
        batch_size=64,
        num_workers=1,
    ):
        """
        After once an twice length Training Dataset was prepared with self._crop_trn_random(),
        this function fills such a dataset into a data loader with perturbation,
        which is also called as sliding-window data augumentation. This is rendered by
        RandomCropepr passed as a collate_fn, leading batched EEG samples to be cropped
        to their half, target length.
        """
        if self._fill_counter == 0:
            # Creates a twice-long training dataset

            # Determines the numbers to sample from each subject
            samps_trn_all = []
            for sub in subs_trn:
                EEGs_sub = data_all["eeg"][data_all["subject"] == sub]
                _n_samps = [
                    EEGs_sub.iloc[i_row].shape[1] // window_size_pts
                    for i_row in range(len(EEGs_sub))
                ]
                samps = np.sum(_n_samps)
                samps_trn_all.append(samps)
            n_trn_subjs = len(samps_trn_all)
            n_95_percentile = sorted(samps_trn_all)[int(n_trn_subjs * 0.05)]
            print(
                f"\n{n_95_percentile} x EEG segments will be sampled from each subject.\n"
            )

            # Makes Training Dataset, whose EEG samples have twice long duration
            self._ds_tra_twice_long = self._crop_trn_random(
                data_all,
                label2class_dict,
                subs_trn,
                labels_trn,
                n_95_percentile,
                subj_str2int_trn,
                subj_str2int_global,
                2 * window_size_pts,
                dtype_np=dtype_np,
                field_order=field_order,
            )

        RC = RandomCropper(window_size_pts, random_state=self._fill_counter)

        dl_tra = DataLoader(
            self._ds_tra_twice_long,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            drop_last=True,
            worker_init_fn=seed_worker,
            pin_memory=True,
            collate_fn=RC.random_crop,
        )
        return dl_tra

    @staticmethod
    def _crop_trn_random(
        data_all,
        label2class_dict,
        subs_trn,
        labels_trn,
        max_n_EEGs_per_sub,
        subj_str2int_trn,
        subj_str2int_global,
        window_size_pts,
        dtype_np=np.float32,
        transform=None,
        field_order=None,
    ):
        def _crop_signal_sub_random(data_all, sub, crop_len, crop_num, dtype):

            EEGs_sub = data_all["eeg"][data_all["subject"] == sub]
            sample_nums = np.array([x.shape[1] for x in EEGs_sub])
            start_cand = sample_nums - crop_len + 1
            start_cand[start_cand < 0] = 0
            start_cumsum = np.cumsum(start_cand)

            EEG_cropped = []
            for crop_i in range(crop_num):
                start_i = np.random.randint(0, np.sum(start_cand), 1)[0]
                rec_i = np.nonzero(start_i < start_cumsum)
                rec_i = rec_i[0][0]

                if rec_i == 0:
                    data_i = start_i
                else:
                    data_i = start_i - start_cumsum[rec_i - 1]
                tmp_signal = EEGs_sub.values[rec_i][:, data_i : (data_i + crop_len)]
                tmp_signal = tmp_signal[np.newaxis, :, :]  # sample (1) * channel * time

                EEG_cropped.append(tmp_signal)

            EEG_cropped = np.concatenate(EEG_cropped, axis=0).astype(dtype)

            return EEG_cropped

        label_num = len(label2class_dict)
        subs_trn_by_classes = [None for _ in range(label_num)]
        for label_i in range(label_num):
            subs_trn_by_classes[label_i] = [
                x for (i, x) in enumerate(subs_trn) if labels_trn[i] == label_i
            ]
        subs_num_trn_by_classes = np.array([len(x) for x in subs_trn_by_classes])
        tot_crop_num_from_a_class = max_n_EEGs_per_sub * np.min(subs_num_trn_by_classes)

        crop_num_trn_by_classes = [None for _ in range(label_num)]
        for label_i in range(label_num):
            div_num = tot_crop_num_from_a_class // subs_num_trn_by_classes[label_i]
            mod_num = tot_crop_num_from_a_class % subs_num_trn_by_classes[label_i]

            crop_num_trn_by_classes[label_i] = (
                np.ones(subs_num_trn_by_classes[label_i]) * div_num
            )

            crop_num_trn_by_classes[label_i][0:mod_num] = (
                crop_num_trn_by_classes[label_i][0:mod_num] + 1
            )

        _data = data_all.set_index("subject").loc[np.hstack(subs_trn_by_classes)]
        _data_uq = _data[~_data.index.duplicated()]

        # fill nan on age, sex, and MMSE by their median values
        _data_uq = fill_age(_data_uq, "age")
        _data_uq = fill_sex(_data_uq, "sex")
        _data_uq = fill_MMSE(_data_uq, "MMSE")

        # normalizing age, sex, and MMSE
        _data_uq["age_normed"] = transform_sex(_data_uq["age"])
        _data_uq["sex_normed"] = transform_sex(_data_uq["sex"])
        _data_uq["MMSE_normed"] = transform_MMSE(_data_uq["MMSE"])
        _data_uq = _data_uq.reset_index()

        X, T, Sg, Sl, A, G, M = [], [], [], [], [], [], []
        for class_i, subs in enumerate(subs_trn_by_classes):
            crop_nums = deepcopy(crop_num_trn_by_classes[class_i])
            np.random.shuffle(crop_nums)
            for sub_i, sub in enumerate(subs):
                EEG_cropped = _crop_signal_sub_random(
                    data_all,
                    sub,
                    window_size_pts,
                    int(crop_nums[sub_i]),
                    dtype_np,
                )
                X.append(EEG_cropped)
                T.append(np.ones((EEG_cropped.shape[0])) * class_i)
                Sg.append(np.ones((EEG_cropped.shape[0])) * subj_str2int_global[sub])
                Sl.append(np.ones((EEG_cropped.shape[0])) * subj_str2int_trn[sub])
                A.append(
                    np.ones((EEG_cropped.shape[0]))
                    * _data_uq[_data_uq["subject"] == sub]["age_normed"].iloc[0]
                )
                G.append(
                    np.ones((EEG_cropped.shape[0]))
                    * _data_uq[_data_uq["subject"] == sub]["sex_normed"].iloc[0]
                )
                M.append(
                    np.ones((EEG_cropped.shape[0]))
                    * _data_uq[_data_uq["subject"] == sub]["MMSE_normed"].iloc[0]
                )

        X = np.concatenate(X, axis=0)
        T = np.concatenate(T, axis=0).astype(np.int32).tolist()
        Sg = np.concatenate(Sg, axis=0).astype(np.int32).tolist()
        Sl = np.concatenate(Sl, axis=0).astype(np.int32).tolist()
        A = np.concatenate(A, axis=0).astype(np.float32).tolist()
        G = np.concatenate(G, axis=0).astype(np.int32).tolist()
        M = np.concatenate(M, axis=0).astype(np.float32).tolist()

        arrs_list_to_pack = sklearn.utils.shuffle(X, T, Sg, Sl, A, G, M)

        assert np.all([len(arr) for arr in arrs_list_to_pack])
        if field_order is not None:
            arrs_list_to_pack = tuple([arrs_list_to_pack[x] for x in field_order])

        ds = _CustomDataset(arrs_list_to_pack, transform=transform)

        return ds

    @classmethod
    def crop_without_overlap(
        cls,
        data_all,
        subs,
        labels,
        crop_len,
        subj_str2int_local,
        subj_str2int_global,
        dtype_np=np.float32,
        crop_num=None,
        transform=None,
        field_order=None,
        batch_size=64,
        num_workers=1,
    ):
        def _crop_signal_sub_without_overlap(data_all, sub, crop_len, crop_num, dtype):
            EEG_cropped = []
            EEGs_sub = data_all["eeg"][data_all["subject"] == sub]
            for EEG in EEGs_sub:

                sample_count = EEG.shape[1] // crop_len
                tmp_signal = EEG[:, 0 : (crop_len * sample_count)]
                tmp_signal = np.reshape(
                    tmp_signal, (tmp_signal.shape[0], crop_len, sample_count), order="F"
                )
                tmp_signal = np.moveaxis(tmp_signal, -1, 0)  # sample * channel * time
                EEG_cropped.append(tmp_signal)

            EEG_cropped = np.concatenate(EEG_cropped, axis=0).astype(dtype)

            if (crop_num is not None) and (EEG_cropped.shape[0] > crop_num):
                EEG_cropped = EEG_cropped[0:crop_num, :, :]

            return EEG_cropped

        _data = data_all.set_index("subject").loc[np.hstack(subs)]
        _data_uq = _data[~_data.index.duplicated()]

        # fill nan on age, sex, and MMSE by their median values
        _data_uq = fill_age(_data_uq, "age")
        _data_uq = fill_sex(_data_uq, "sex")
        _data_uq = fill_MMSE(_data_uq, "MMSE")

        # normalizing age, sex, and MMSE
        _data_uq["age_normed"] = transform_age(_data_uq["age"])
        _data_uq["sex_normed"] = transform_sex(_data_uq["sex"])
        _data_uq["MMSE_normed"] = transform_MMSE(_data_uq["MMSE"])
        _data_uq = _data_uq.reset_index()

        X, T, Sg, Sl, A, G, M = [], [], [], [], [], [], []
        for sub_i, sub in enumerate(subs):
            EEG_cropped = _crop_signal_sub_without_overlap(
                data_all, sub, crop_len, crop_num, dtype_np
            )
            X.append(EEG_cropped)
            T.append(np.ones((EEG_cropped.shape[0])) * labels[sub_i])
            Sg.append(np.ones((EEG_cropped.shape[0])) * subj_str2int_global[sub])
            Sl.append(np.ones((EEG_cropped.shape[0])) * subj_str2int_local[sub])
            A.append(
                np.ones((EEG_cropped.shape[0]))
                * _data_uq[_data_uq["subject"] == sub]["age_normed"].iloc[0]
            )
            G.append(
                np.ones((EEG_cropped.shape[0]))
                * _data_uq[_data_uq["subject"] == sub]["sex_normed"].iloc[0]
            )
            M.append(
                np.ones((EEG_cropped.shape[0]))
                * _data_uq[_data_uq["subject"] == sub]["MMSE_normed"].iloc[0]
            )

        assert (
            len(subj_str2int_local)
            == len(np.unique(np.hstack(Sg)))
            == len(np.unique(np.hstack(Sl)))
        )
        assert set(subj_str2int_local.values()) == set(np.unique(np.hstack(Sl)))

        X = np.concatenate(X, axis=0)
        T = np.concatenate(T, axis=0).astype(np.int32).tolist()
        Sg = np.concatenate(Sg, axis=0).astype(np.int32).tolist()
        Sl = np.concatenate(Sl, axis=0).astype(np.int32).tolist()
        A = np.concatenate(A, axis=0).astype(np.float32).tolist()
        G = np.concatenate(G, axis=0).astype(np.int32).tolist()
        M = np.concatenate(M, axis=0).astype(np.float32).tolist()

        arrs_list_to_pack = sklearn.utils.shuffle(X, T, Sg, Sl, A, G, M)
        assert np.all([len(arr) for arr in arrs_list_to_pack])
        if field_order is not None:
            arrs_list_to_pack = tuple([arrs_list_to_pack[x] for x in field_order])

        ds = _CustomDataset(
            arrs_list_to_pack,
            transform=transform,
            order=field_order,
        )
        # Please allow "shuffle=True" to estimate validation metrics from only batches
        dl = DataLoader(
            ds,
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            drop_last=False,
            worker_init_fn=seed_worker,
            pin_memory=True,
        )
        return dl

    @staticmethod
    def _fix_seeds(random_state=42):
        random.seed(random_state)
        np.random.seed(random_state)
        torch.manual_seed(random_state)
        torch.cuda.manual_seed(random_state)
        torch.cuda.manual_seed_all(random_state)
        torch.backends.cudnn.deterministic = True

    @staticmethod
    def _load_and_process_filler_params(conc_classes):
        # loads filler_params
        filler_params = mngs.io.load("./config/filler_params.yaml")
        filler_params["conc_classes"] = conc_classes
        filler_params.setdefault("class_field_name", "disease_type")
        # filler_params.setdefault("window_size_pts", 1000)
        filler_params.setdefault("random_state", 42)
        filler_params.setdefault("batch_size", 64)
        filler_params.setdefault("num_workers", 16)
        filler_params.setdefault("dtype", "fp32")
        filler_params.setdefault("field_order", [0, 1, 2, 3, 4, 5, 6])
        filler_params.setdefault("transform", None)
        filler_params.setdefault("is_debug_mode", False)

        # additional params
        filler_params["_dtype_np"] = (
            np.float32 if filler_params["dtype"] == "fp32" else np.float16
        )

        return filler_params

    @classmethod
    def _check_cv_health(cls, cv_dict, i_fold):
        assert len(set(cv_dict["subs_trn"][i_fold])) == len(cv_dict["subs_trn"][i_fold])
        assert len(set(cv_dict["subs_val"][i_fold])) == len(cv_dict["subs_val"][i_fold])
        assert len(set(cv_dict["subs_tes"][i_fold])) == len(cv_dict["subs_tes"][i_fold])

        assert (
            len(set(cv_dict["subs_trn"][i_fold]) & set(cv_dict["subs_val"][i_fold]))
            == 0
        )
        assert (
            len(set(cv_dict["subs_val"][i_fold]) & set(cv_dict["subs_tes"][i_fold]))
            == 0
        )
        assert (
            len(set(cv_dict["subs_tes"][i_fold]) & set(cv_dict["subs_trn"][i_fold]))
            == 0
        )
        print("\nCross Validation dict is healthy.\n")

    @staticmethod
    def _mk_db(data_all, cv_dict):
        """Makes a global database."""
        db = pd.DataFrame()
        db = (
            data_all[
                ["subject", "disease_type", "cognitive_level", "age", "sex", "MMSE"]
            ]
            .drop_duplicates()
            .set_index("subject")
        )
        db = db.rename(columns={"disease_type": "disease_type_str"})
        disease_type_int = []

        for ii in range(len(db)):
            try:
                disease_type_int.append(
                    cv_dict["conc_class_str_2_label_int_dict"][
                        db["disease_type_str"].iloc[ii]
                    ]
                )
            except Exception as e:
                print(e)
                disease_type_int.append(-1)

        db["disease_type_int"] = disease_type_int
        db["global_id"] = np.arange(len(db))

        for i_fold, (subs_trn, subs_val, subs_tes) in enumerate(
            zip(cv_dict["subs_trn"], cv_dict["subs_val"], cv_dict["subs_tes"])
        ):
            db.loc[subs_trn, f"fold#{i_fold}_dataset"] = "Training"
            db.loc[subs_trn, f"fold#{i_fold}_local_id"] = np.arange(
                len(subs_trn), dtype=int
            )

            db.loc[subs_val, f"fold#{i_fold}_dataset"] = "Validation"
            db.loc[subs_val, f"fold#{i_fold}_local_id"] = np.arange(
                len(subs_val), dtype=int
            )

            db.loc[subs_tes, f"fold#{i_fold}_dataset"] = "Test"
            db.loc[subs_tes, f"fold#{i_fold}_local_id"] = np.arange(
                len(subs_tes), dtype=int
            )
        return db

    @classmethod
    def load_BIDS(
        cls,
        BIDS_ROOT,
        disease_types=["HV", "AD", "NPH", "DLB"],
        samp_rate=500,
        drop_cMCI=False,
        montage=None,
    ):

        load_params = mngs.io.load("./config/load_params.yaml")

        if montage != None:
            load_params["montage"] = montage

        load_params["BIDS_root"] = BIDS_ROOT
        load_params["conc_classes"] = disease_types  # e.g., ["AD", "DLB", "HV", "NPH"]
        dataset_str = BIDS_ROOT.split("/")[-1]

        data_all = load_BIDS_func(load_params)                

        ## drops not-labeled? data
        expanded_classes = np.hstack(_expand_conc_classes_str(disease_types))
        indi_to_use = (
            pd.concat(
                [data_all["disease_type"] == ec for ec in expanded_classes], axis=1
            )
            .sum(axis=1)
            .astype(bool)
        )
        if drop_cMCI:
            indi_drop = data_all["cognitive_level"] == "cMCI"
            indi_to_use = indi_to_use * ~indi_drop

        data_not_used = data_all.loc[~indi_to_use]
        data_all = data_all.loc[indi_to_use]

        return load_params, data_all, data_not_used

    @property
    def sex_str2int_dict(self):
        return {"M": 0, "F": 1}

    @property
    def sex_int2str_dict(self):
        return {v: k for k, v in self.sex_str2int_dict.items()}

    @property
    def sub_str2int_global(
        self,
    ):
        return {s: i for s, i in zip(self.db.index, self.db["global_id"])}

    @property
    def sub_int2str_global(
        self,
    ):
        return {i: s for s, i in self.sub_str2int_global.items()}

    def _sub_str2int_local(self, dataset_str):
        _db = self.db[self.db[f"fold#{self.i_fold}_dataset"] == dataset_str][
            f"fold#{self.i_fold}_local_id"
        ]
        return {s: int(i) for s, i in zip(_db.index, _db)}

    def _sub_int2str_local(self, dataset_str):
        return {i: s for s, i in self._sub_str2int_local(dataset_str).items()}

    @property
    def sub_str2int_trn(
        self,
    ):
        return self._sub_str2int_local("Training")

    @property
    def sub_int2str_trn(
        self,
    ):
        return self._sub_int2str_local("Training")

    @property
    def sub_str2int_val(
        self,
    ):
        return self._sub_str2int_local("Validation")

    @property
    def sub_int2str_val(
        self,
    ):
        return self._sub_int2str_local("Validation")

    @property
    def sub_str2int_tes(
        self,
    ):
        return self._sub_str2int_local("Test")

    @property
    def sub_int2str_tes(
        self,
    ):
        return self._sub_int2str_local("Test")

    @property
    def subs_all(
        self,
    ):
        return np.array(
            natsorted(np.hstack([self.subs_trn, self.subs_val, self.subs_tes]))
        )

    @property
    def subs_trn(
        self,
    ):
        return list(self.sub_int2str_trn.values())

    @property
    def subs_val(
        self,
    ):
        return list(self.sub_int2str_val.values())

    @property
    def subs_tes(
        self,
    ):
        return list(self.sub_int2str_tes.values())

    def _labels_dataset(self, dataset_str):
        return np.array(
            self.db["disease_type_int"][
                self.db[f"fold#{self.i_fold}_dataset"] == dataset_str
            ]
        )

    @property
    def labels_trn(
        self,
    ):
        return self._labels_dataset("Training")

    @property
    def labels_val(
        self,
    ):
        return self._labels_dataset("Validation")

    @property
    def labels_tes(
        self,
    ):
        return self._labels_dataset("Test")

    @classmethod
    def transform_age(cls, x):
        return transform_age(x)

    @classmethod
    def transform_sex(cls, x):
        return transform_sex(x)

    @classmethod
    def transform_MMSE(cls, x):
        return transform_MMSE(x)


class RandomCropper:
    def __init__(self, window_size_pts, random_state=42):
        self.window_size_pts = window_size_pts
        # self.random_state = random_state
        self.rs = np.random.RandomState(random_state)

    def random_crop(self, data_list):
        """
        len(data_list) # batch_size
        tensors_list = data_list[0]
        len(tensors_list) # 4 # Xb, Tb, Sgb, Slb

        A transform function on batched EEG segments.
        Note: X_arr.shape -> [n_chs, window_size_pts]
        """

        n_elements = len(data_list[0])

        elements = []
        for i_elements in range(n_elements):
            elements.append(torch.stack([tl[i_elements] for tl in data_list]))

        # Randomly crop EEG segments
        Xb = elements[0]
        rand_start = self.rs.randint(Xb.shape[-1] - (self.window_size_pts - 1))
        rand_end = rand_start + self.window_size_pts
        Xb = Xb[..., rand_start:rand_end]
        elements[0] = Xb

        # print(self.random_state)

        return elements


class _CustomDataset(Dataset):
    """
    Example:
        n = 1024
        n_chs = 19
        X = np.random.rand(n, n_chs, 1000)
        T = np.random.randint(0, 4, size=(n, 1))
        S = np.random.randint(0, 999, size=(n, 1))
        Sl = np.random.randint(0, 4, size=(n, 1))

        arrs_list = [X, T, S, Sl]
        transform = None
        ds = _CustomDataset(arrs_list, transform=transform)
        len(ds) # 1024
    """

    def __init__(self, arrs_list, transform=None, order=[0, 1, 2, 3, 4, 5, 6]):
        """
        Arguments:
            'transform'
                a function for transforming X (= EEG) like applying data augmentation.
                Pseudo-code is something like below.
                X_arr = transform(X_arr) # X_arr.shape [n_chs, window_size_pts]
        """

        self.arrs_list = arrs_list
        assert np.all([len(arr) for arr in arrs_list])
        self.length = len(arrs_list[0])
        self.transform = transform
        self.order = order

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        r_list = [l[idx] for l in self.arrs_list]

        if self.transform:
            dtype_orig = r_list[0].dtype
            r_list[self.order.index(0)] = self.transform(
                r_list[0].astype(np.float64)
            ).astype(dtype_orig)

        return [torch.tensor(d) for d in r_list]


def seed_worker(worker_id):
    """
    https://dajiro.com/entry/2021/04/13/233032
    """
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def transform_age(x):
    x = x.copy()
    return (x - x.mean()) / x.std()


def transform_sex(x):
    x = x.copy()
    return x


def transform_MMSE(x):
    x = x.copy()
    minus_MMSE = np.array(31 - x)
    log_minus_MMSE = np.log(minus_MMSE + 1e-5)
    norm_log_minus_MMSE = (
        log_minus_MMSE - log_minus_MMSE.mean()
    ) / log_minus_MMSE.std()
    return norm_log_minus_MMSE


def fill_age(df, key):
    return _fill_by_median(df, key)


def fill_MMSE(df, key):
    return _fill_by_median(df, key)


def _fill_by_median(df, key):
    col = df[key].copy()
    median = np.nanmedian(col)
    df.loc[col.isna(), key] = median
    return df


def fill_sex(df, key):
    col = df[key].copy()
    val = 0.5
    df.loc[col.isna(), key] = val
    return df


if __name__ == "__main__":
    disease_types = ["HV", "AD", "NPH", "DLB"]
    dlf = DataLoaderFiller(
        "./data/BIDS_Kochi",
        disease_types,
        drop_cMCI=True,
    )
    dlf.fill(i_fold=0, reset_fill_counter=True)

    """    
    SG_tra = dlf.dl_tra.dataset.arrs_list[2]
    SG_val = dlf.dl_val.dataset.arrs_list[2]
    SG_tes = dlf.dl_tes.dataset.arrs_list[2]
    len(np.unique(SG_tra + SG_val + SG_tes)) # 204

    T_tra = dlf.dl_tra.dataset.arrs_list[1]
    T_val = dlf.dl_val.dataset.arrs_list[1]
    T_tes = dlf.dl_tes.dataset.arrs_list[1]
    np.unique(T_tra, return_counts=True)
    np.unique(np.array(T_tra + T_val + T_tes), return_counts=True)
    
    """

    n_epochs = 3
    for i_fold in range(dlf.load_params["num_folds"]):
        dlf.fill(i_fold, reset_fill_counter=True)
        for epoch in range(n_epochs):
            dlf.fill(i_fold, reset_fill_counter=False)

            # Training dataset
            for i_batch, batch in enumerate(dlf.dl_tra):
                Xb_tra, Tb_tra, Sgb_tra, Slb_tra, A_tra, G_tra, M_tra = batch
                print(Xb_tra)

            # Validation dataset
            for i_batch, batch in enumerate(dlf.dl_val):
                Xb_val, Tb_val, Sgb_val, Slb_val, A_val, G_val, M_val = batch

        # Test dataset
        for i_batch, batch in enumerate(dlf.dl_tra):
            Xb_tra, Tb_tra, Sgb_tra, Slb_tra, A_tra, G_tra, M_tra = batch



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import mngs

################################################################################
## MNIST
################################################################################
from sklearn import datasets, metrics, svm
from sklearn.model_selection import train_test_split

digits = datasets.load_digits()

# flatten the images
n_samples = len(digits.images)
data = digits.images.reshape((n_samples, -1))

# Create a classifier: a support vector classifier
clf = svm.SVC(gamma=0.001, probability=True)

# Split data into 50% train and 50% test subsets
X_train, X_test, y_train, y_test = train_test_split(
    data, digits.target, test_size=0.5, shuffle=False
)

# Learn the digits on the train subset
clf.fit(X_train, y_train)

# Predict the value of the digit on the test subset
predicted_proba = clf.predict_proba(X_test)
predicted = clf.predict(X_test)

n_classes = len(np.unique(digits.target))
labels = ["Class {}".format(i) for i in range(n_classes)]

## Configures matplotlib
plt.rcParams["font.size"] = 20
plt.rcParams["legend.fontsize"] = "xx-small"
scale = 0.75
plt.rcParams["figure.figsize"] = (16 * scale, 9 * scale)


################################################################################
## Main
################################################################################
## ROC Curve
fig_roc, metrics_roc = mngs.ml.plt.roc_auc(plt, y_test, predicted_proba, labels)
fig_roc.show()
## Precision-Recall Curve
fig_pre_rec, metrics_pre_rec = mngs.ml.plt.pre_rec_auc(
    plt, y_test, predicted_proba, labels
)
fig_pre_rec.show()

## EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-03-17 21:30:11 (ywatanabe)"

import numpy as np
from scipy.linalg import norm

def cosine(v1, v2):
    if np.isnan(v1).any():
        return np.nan
    if np.isnan(v2).any():
        return np.nan
    return np.dot(v1, v2) / (norm(v1) * norm(v2))

def nannorm(v, axis=-1):
    if np.isnan(v).any():
        return np.nan
    else:
        return norm(v, axis=axis)

def rebase_a_vec(v, v_base):
    def production_vector(v1, v0):
        """
        production_vector(np.array([3,4]), np.array([10,0])) # np.array([3, 0])
        """
        return norm(v1) * cosine(v1, v0) * v0 / norm(v0)

    if np.isnan(v).any():
        return np.nan
    if np.isnan(v_base).any():
        return np.nan
    v_prod = production_vector(v, v_base)
    sign = np.sign(cosine(v, v_base))
    return sign * norm(v_prod)

import sympy
import numpy as np


def three_line_lengths_to_coords(aa, bb, cc):
    """    
    O, A, B = three_line_lengths_to_coords(2, np.sqrt(3), 1)
    print(O, A, B)
    """
    
    # Definition
    a1 = sympy.Symbol("a1")    
    b1 = sympy.Symbol("b1")
    b2 = sympy.Symbol("b2")    
    
    a1 = aa
    # b1 = bb

    # Calculates
    cos = (aa**2 + bb**2 - cc**2) / (2 * aa * bb)
    sin = np.sqrt(1 - cos**2)
    S1 = 1 / 2 * aa * bb * sin
    S2 = 1 / 2 * aa * b2

    # Solves
    b2 = sympy.solve(S1 - S2)[0]
    b1 = bb * cos

    # tan1 = b2 / b1
    # tan2 = sin/cos

    # b1 = sympy.solve(tan1-tan2)[0]
    O = (0, 0, 0)
    A = (a1, 0, 0)
    B = (b1, b2, 0)

    return O, A, B



#! ./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-16 17:07:27"


"""
This script does XYZ.
"""

# Imports
import sys

import matplotlib.pyplot as plt
import mngs
import numpy as np
import tensorpac


# Functions
def calc_pac_with_tensorpac(xx, fs, t_sec, i_batch=0, i_ch=0):
    # Morlet's Wavelet Transfrmation
    p = tensorpac.Pac(f_pha="hres", f_amp="mres", dcomplex="wavelet")

    # Bandpass Filtering and Hilbert Transformation
    phases = p.filter(
        fs, xx[i_batch, i_ch], ftype="phase", n_jobs=1
    )  # (50, 20, 2048)
    amplitudes = p.filter(
        fs, xx[i_batch, i_ch], ftype="amplitude", n_jobs=1
    )  # (50, 20, 2048)

    # Calculates xpac
    k = 2
    p.idpac = (k, 0, 0)
    xpac = p.fit(phases, amplitudes)  # (50, 50, 20)
    pac = xpac.mean(axis=-1)  # (50, 50)

    freqs_amp = p.f_amp.mean(axis=-1)
    freqs_pha = p.f_pha.mean(axis=-1)

    pac = pac.T  # (amp, pha) -> (pha, amp)

    return phases, amplitudes, freqs_pha, freqs_amp, pac


def plot_PAC_mngs_vs_tensorpac(pac_mngs, pac_tp, freqs_pha, freqs_amp):
    assert pac_mngs.shape == pac_tp.shape

    # Plots
    fig, axes = mngs.plt.subplots(ncols=3)  # , sharex=True, sharey=True

    # To align scalebars
    vmin = min(np.min(pac_mngs), np.min(pac_tp), np.min(pac_mngs - pac_tp))
    vmax = max(np.max(pac_mngs), np.max(pac_tp), np.max(pac_mngs - pac_tp))

    # mngs version
    ax = axes[0]
    ax.imshow2d(
        pac_mngs,
        cbar=False,
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_title("mngs")

    # Tensorpac
    ax = axes[1]
    ax.imshow2d(
        pac_tp,
        cbar=False,
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_title("Tensorpac")

    # Diff.
    ax = axes[2]
    ax.imshow2d(
        pac_mngs - pac_tp,
        cbar_label="PAC values",
        cbar_shrink=0.5,
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_title(f"Difference\n(mngs - Tensorpac)")

    # for ax in axes:
    #     ax.set_ticks(
    #         x_vals=freqs_pha,
    #         # y_vals=freqs_amp,
    #     )
    #     # ax.set_n_ticks()

    fig.suptitle("PAC (MI) values")
    fig.supxlabel("Frequency for phase [Hz]")
    fig.supylabel("Frequency for amplitude [Hz]")

    return fig


if __name__ == "__main__":
    import torch

    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # Parameters
    FS = 512
    T_SEC = 4

    xx, tt, fs = mngs.dsp.demo_sig(
        batch_size=2,
        n_chs=2,
        n_segments=2,
        fs=FS,
        t_sec=T_SEC,
        sig_type="tensorpac",
    )

    # mngs
    pac_mngs, freqs_pha, freqs_amp = mngs.dsp.pac(
        xx, fs, batch_size=2, pha_n_bands=50, amp_n_bands=30
    )
    i_batch, i_epoch = 0, 0
    pac_mngs = pac_mngs[i_batch, i_epoch]

    # Tensorpac
    phases, amplitudes, freqs_pha, freqs_amp, pac_tp = calc_pac_with_tensorpac(
        xx, fs, T_SEC, i_batch=0, i_ch=0
    )

    # Plots
    fig = plot_PAC_mngs_vs_tensorpac(pac_mngs, pac_tp, freqs_pha, freqs_amp)
    plt.show()

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/dsp/utils/pac.py
"""



#!/usr/bin/env python3
import matplotlib
import mngs
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable
from sklearn.metrics import confusion_matrix as sklearn_confusion_matrix


def conf_mat(
    plt,
    cm=None,
    y_true=None,
    y_pred=None,
    y_pred_proba=None,
    labels=None,
    sorted_labels=None,
    pred_labels=None,
    sorted_pred_labels=None,
    true_labels=None,
    sorted_true_labels=None,
    label_rotation_xy=(15, 15),
    title=None,
    colorbar=True,
    x_extend_ratio=1.0,
    y_extend_ratio=1.0,
    spath=None,
):
    """
    Inverse the y-axis and plot the confusion matrix as a heatmap.
    The predicted labels (in x-axis) is symboled with hat (^).
    The plt object is passed to adjust the figure size

    cm = sklearn.metrics.confusion_matrix(y_test, y_pred)


    cm = np.random.randint(low=0, high=10, size=[3,4])
    x: predicted labels
    y: true_labels


    kwargs:

        "extend_ratio":
            Determines how much the axes objects (not the fig object) are extended
            in the vertical direction.

    """

    if (y_pred_proba is not None) and (y_pred is None):
        y_pred = y_pred_proba.argmax(axis=-1)

    assert (cm is not None) or ((y_true is not None) and (y_pred is not None))

    if not cm:
        cm = sklearn_confusion_matrix(y_true, y_pred)

    # Dataframe
    df = pd.DataFrame(data=cm).copy()

    # To LaTeX styles
    if pred_labels is not None:
        pred_labels = [mngs.general.to_the_latex_style(l) for l in pred_labels]
    if true_labels is not None:
        true_labels = [mngs.general.to_the_latex_style(l) for l in true_labels]
    if labels is not None:
        labels = [mngs.general.to_the_latex_style(l) for l in labels]
    if sorted_labels is not None:
        sorted_labels = [
            mngs.general.to_the_latex_style(l) for l in sorted_labels
        ]

    # Prediction Labels: columns
    if pred_labels is not None:
        df.columns = pred_labels
    elif (pred_labels is None) and (labels is not None):
        df.columns = labels

    # Ground Truth Labels: index
    if true_labels is not None:
        df.index = true_labels
    elif (true_labels is None) and (labels is not None):
        df.index = labels

    # Sort based on sorted_labels here
    if sorted_labels is not None:
        assert set(sorted_labels) == set(labels)
        df = df.reindex(index=sorted_labels, columns=sorted_labels)

    # Main
    fig, ax = plt.subplots()
    res = sns.heatmap(
        df,
        annot=True,
        annot_kws={"size": plt.rcParams["font.size"]},
        fmt=".0f",
        cmap="Blues",
        cbar=False,
    )  # Here, don't plot color bar.

    # Adds comma separator for the annotated int texts
    for t in ax.texts:
        t.set_text("{:,d}".format(int(t.get_text())))

    # Inverts the y-axis
    res.invert_yaxis()

    # Makes the frame visible
    for _, spine in res.spines.items():
        spine.set_visible(False)

    # Labels
    ax.set_xlabel("Predicted label")
    ax.set_ylabel("True label")
    ax.set_title(title)

    # Appearances
    ax = mngs.plt.ax.extend(ax, x_extend_ratio, y_extend_ratio)
    if df.shape[0] == df.shape[1]:
        ax.set_box_aspect(1)
    ax.set_xticklabels(
        ax.get_xticklabels(),
        rotation=label_rotation_xy[0],
        fontdict={"verticalalignment": "top"},
    )
    ax.set_yticklabels(
        ax.get_yticklabels(),
        rotation=label_rotation_xy[1],
        fontdict={"horizontalalignment": "right"},
    )
    # The size
    bbox = ax.get_position()
    left_orig = bbox.x0
    width_orig = bbox.x1 - bbox.x0
    g_x_orig = left_orig + width_orig / 2.0
    width_tgt = width_orig * x_extend_ratio  # x_extend_ratio
    dx = width_orig - width_tgt

    # Adjusts the sizes of the confusion matrix and colorbar
    if colorbar == True:  # fixme
        divider = make_axes_locatable(ax)  # Gets region from the ax
        cax = divider.append_axes("right", size="5%", pad=0.1)
        # cax = divider.new_horizontal(size="5%", pad=1, pack_start=True)
        cax = mngs.plt.ax.set_pos(fig, cax, -dx * 2.54, 0)
        fig.add_axes(cax)

        """
        axpos = ax.get_position()
        caxpos = cax.get_position()

        AddAxesBBoxRect(fig, ax, ec="r")
        AddAxesBBoxRect(fig, cax, ec="b")

        fig.text(
            axpos.x0 + 0.01, axpos.y0 + 0.01, "after colorbar", weight="bold", color="r"
        )

        fig.text(
            caxpos.x1 + 0.01,
            caxpos.y1 - 0.01,
            "cax position",
            va="top",
            weight="bold",
            color="b",
            rotation="vertical",
        )
        """

        # Plots colorbar and adjusts the size
        vmax = np.array(df).max().astype(int)
        norm = matplotlib.colors.Normalize(vmin=0, vmax=vmax)
        cbar = fig.colorbar(
            plt.cm.ScalarMappable(norm=norm, cmap="Blues"),
            cax=cax,
            # shrink=0.68,
        )
        cbar.locator = ticker.MaxNLocator(nbins=4)  # tick_locator
        cbar.update_ticks()
        # cbar.outline.set_edgecolor("#f9f2d7")
        cbar.outline.set_edgecolor("white")

    if spath is not None:
        mngs.io.save(fig, spath)

    return fig, cm


# def AddAxesBBoxRect(fig, ax, ec="k"):
#     from matplotlib.patches import Rectangle

#     axpos = ax.get_position()
#     rect = fig.patches.append(
#         Rectangle(
#             (axpos.x0, axpos.y0),
#             axpos.width,
#             axpos.height,
#             ls="solid",
#             lw=2,
#             ec=ec,
#             fill=False,
#             transform=fig.transFigure,
#         )
#     )
#     return rect


if __name__ == "__main__":

    import mngs

    y_true, y_pred = mngs.io.load("/tmp/tmp.pkl")

    fig, cm = conf_mat(plt, y_true=y_true, y_pred=y_pred)

    # https://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html#sphx-glr-auto-examples-model-selection-plot-confusion-matrix-py
    import sys

    import matplotlib.pyplot as plt
    import numpy as np
    import sklearn
    from sklearn import datasets, svm

    # from sklearn.metrics import plot_confusion_matrix
    from sklearn.model_selection import train_test_split

    sys.path.append(".")
    import mngs

    # Imports some data to play with
    iris = datasets.load_iris()
    X = iris.data
    y = iris.target
    class_names = iris.target_names

    # Splits the data into a training set and a test set
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    # Runs classifier, using a model that is too regularized (C too low) to see
    # the impact on the results
    classifier = svm.SVC(kernel="linear", C=0.01).fit(X_train, y_train)

    ## Checks the confusion_matrix function
    y_pred = classifier.predict(X_test)
    cm = sklearn.metrics.confusion_matrix(y_test, y_pred)
    cm **= 3

    cm = np.random.randint(low=0, high=10, size=[3, 4])

    mngs.plt.configure_mpl(
        plt,
        # figsize=(4, 8),
        # figsize=(4, 8),
        # fontsize=6,
        # labelsize=8,
        # legendfontsize=7,
        # tick_size=0.8,
        # tick_width=0.2,
    )

    # labels = class_names
    pred_labels = ["A", "B", "C"]
    true_labels = ["a", "b", "c", "d"]

    fig, cm = conf_mat(
        plt,
        cm,
        # labels=class_names,
        pred_labels=pred_labels,
        true_labels=true_labels,
        label_rotation_xy=(60, 60),
        x_extend_ratio=1.0,
        colorbar=True,
    )

    fig.axes[-1] = mngs.plt.ax.sci_note(
        fig.axes[-1],
        3,
        fformat="%3.1f",
        # fformat="%d",
        y=True,
    )

    fig.show()

    ## EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-17 08:37:31"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

# Imports
import matplotlib.pyplot as plt


# Functions
def set_supxyt(ax, xlabel=None, ylabel=None, title=None):
    fig = ax.get_figure()
    """Sets xlabel, ylabel and title"""
    if xlabel is not None:
        fig.supxlabel(xlabel)
    if ylabel is not None:
        fig.supylabel(ylabel)
    if title is not None:
        fig.suptitle(title)
    return ax


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/plt/ax/_set_lt.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-02-03 16:00:30 (ywatanabe)"

from ._SigMacro_processFigure_S import SigMacro_processFigure_S
from ._SigMacro_toBlue import SigMacro_toBlue



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-07 19:14:40 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/_subplots/AxesWrapper.py

import pandas as pd


class AxesWrapper:
    """
    A wrapper class for multiple axes objects that can convert each to a pandas DataFrame
    and concatenate them into a single DataFrame.
    """

    def __init__(self, axes):
        """
        Initializes the AxesWrapper with a list of axes.

        Parameters:
            axes (iterable): An iterable of objects that have a to_sigma() method returning a DataFrame.
        """

        self.axes = axes
        # self.shape = (len(axes),)  # [REVISED]

    def __getitem__(self, index):
        """
        Allows the AxesWrapper to be indexed.

        Parameters:
            index (int or slice): The index or slice of the axes to retrieve.

        Returns:
            The axis at the specified index or a new AxesWrapper with the sliced axes.
        """
        if isinstance(index, slice):
            # Return a new AxesWrapper object containing the sliced part
            return AxesWrapper(self.axes[index])
        else:
            # Return the specific axis at the index
            return self.axes[index]

    def __iter__(self):
        return iter(self.axes)

    def __len__(self):
        return len(self.axes)

    @property
    def shape(self):
        return self.axes.shape

    def to_sigma(self):
        """
        Converts each axis to a DataFrame using their to_sigma method and concatenates them along columns.

        Returns:
            DataFrame: A concatenated DataFrame of all axes.
        """
        dfs = []
        for i_ax, ax in enumerate(self.axes.ravel()):
            df = ax.to_sigma()
            df.columns = [f"{i_ax}_{col}" for col in df.columns]
            dfs.append(df)
        return pd.concat(dfs, axis=1)

    def ravel(self):
        """
        Flattens the AxesWrapper into a 1D array-like structure of axes.

        Returns:
            list: A list containing all axes objects.
        """
        self.axes = self.axes.ravel()  # [REVISED]
        return self.axes

    def flatten(self):
        """
        Flattens the AxesWrapper into a 1D array-like structure of axes.

        Returns:
            list: A list containing all axes objects.
        """
        self.axes = self.axes.flatten()
        return self.axes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-05-15 17:09:58 (ywatanabe)"

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchsummary import summary
import mngs
import numpy as np
import mngs


class BHead(nn.Module):
    def __init__(self, n_chs_in, n_chs_out):
        super().__init__()
        self.sa = mngs.nn.SpatialAttention(n_chs_in)
        self.conv11 = nn.Conv1d(
            in_channels=n_chs_in, out_channels=n_chs_out, kernel_size=1
        )

    def forward(self, x):
        x = self.sa(x)
        x = self.conv11(x)
        return x


class BNet(nn.Module):
    def __init__(self, BNet_config, MNet_config):
        super().__init__()
        self.dummy_param = nn.Parameter(torch.empty(0))
        # N_VIRTUAL_CHS = 32
        # "n_virtual_chs":16,        

        self.sc = mngs.nn.SwapChannels()
        self.dc = mngs.nn.DropoutChannels(dropout=0.01)        
        self.fgc = mngs.nn.FreqGainChanger(
            BNet_config["n_bands"], BNet_config["SAMP_RATE"]
        )
        self.heads = nn.ModuleList(
            [
                BHead(n_ch, BNet_config["n_virtual_chs"]).to(self.dummy_param.device)
                for n_ch in BNet_config["n_chs_of_modalities"]
            ]
        )

        self.cgcs = [mngs.nn.ChannelGainChanger(n_ch) for n_ch in BNet_config["n_chs_of_modalities"]]
        # self.cgc = mngs.nn.ChannelGainChanger(N_VIRTUAL_CHS)        

        # MNet_config["n_chs"] = BNet_config["n_virtual_chs"]  # BNet_config["n_chs"] # override

        n_chs = BNet_config["n_virtual_chs"]
        self.blk1 = mngs.nn.ResNetBasicBlock(n_chs, n_chs)
        self.blk2 = mngs.nn.ResNetBasicBlock(int(n_chs/2**1), int(n_chs/2**1))
        self.blk3 = mngs.nn.ResNetBasicBlock(int(n_chs/2**2), int(n_chs/2**2))
        self.blk4 = mngs.nn.ResNetBasicBlock(int(n_chs/2**3), int(n_chs/2**3))
        self.blk5 = mngs.nn.ResNetBasicBlock(1, 1)
        self.blk6 = mngs.nn.ResNetBasicBlock(1, 1)
        self.blk7 = mngs.nn.ResNetBasicBlock(1, 1)

        # self.MNet = mngs.nn.MNet_1000(MNet_config)

        # self.fcs = nn.ModuleList(
        #     [
        #         nn.Sequential(
        #             # nn.Linear(N_FC_IN, config["n_fc1"]),
        #             nn.Mish(),
        #             nn.Dropout(BNet_config["d_ratio1"]),
        #             nn.Linear(BNet_config["n_fc1"], BNet_config["n_fc2"]),
        #             nn.Mish(),
        #             nn.Dropout(BNet_config["d_ratio2"]),
        #             nn.Linear(BNet_config["n_fc2"], BNet_config["n_classes_of_modalities"][i_head]),
        #         )
        #         for i_head, _ in enumerate(range(len(BNet_config["n_chs_of_modalities"])))
        #     ]
        # )

    @staticmethod
    def _znorm_along_the_last_dim(x):
        return (x - x.mean(dim=-1, keepdims=True)) / x.std(dim=-1, keepdims=True)

    def forward(self, x, i_head):
        x = self._znorm_along_the_last_dim(x)        
        # x = self.sc(x)
        x = self.dc(x)        
        x = self.fgc(x)
        x = self.cgcs[i_head](x)
        x = self.heads[i_head](x)

        x = self.blk1(x)
        x = F.avg_pool1d(x.transpose(1,2), kernel_size=2).transpose(1,2)
        x = F.avg_pool1d(x, kernel_size=2)
        x = self.blk2(x)
        x = F.avg_pool1d(x.transpose(1,2), kernel_size=2).transpose(1,2)
        x = F.avg_pool1d(x, kernel_size=2)        
        x = self.blk3(x)
        x = F.avg_pool1d(x.transpose(1,2), kernel_size=2).transpose(1,2)
        x = F.avg_pool1d(x, kernel_size=2)        
        x = self.blk4(x)
        x = F.avg_pool1d(x.transpose(1,2), kernel_size=2).transpose(1,2)
        x = F.avg_pool1d(x, kernel_size=2)        

        x = self.blk5(x)
        x = F.avg_pool1d(x, kernel_size=2)        
        x = self.blk6(x)
        x = F.avg_pool1d(x, kernel_size=2)        
        x = self.blk7(x)
        x = F.avg_pool1d(x, kernel_size=2)        


        import ipdb; ipdb.set_trace()                        


        # x = self.cgc(x)
        x = self.MNet.forward_bb(x)
        x = self.fcs[i_head](x)
        return x


# BNet_config = {
#     "n_chs": 32,
#     "n_bands": 6,
#     "SAMP_RATE": 1000,
# }
BNet_config = {
    "n_bands": 6,
    "n_virtual_chs":16,
    "SAMP_RATE": 250,
    "n_fc1": 1024,
    "d_ratio1": 0.85,
    "n_fc2": 256,
    "d_ratio2": 0.85,
}


if __name__ == "__main__":
    ## Demo data
    # MEG
    BS, N_CHS, SEQ_LEN = 16, 160, 1024
    x_MEG = torch.rand(BS, N_CHS, SEQ_LEN).cuda()
    # EEG
    BS, N_CHS, SEQ_LEN = 16, 19, 1024
    x_EEG = torch.rand(BS, N_CHS, SEQ_LEN).cuda()

    
    # m = mngs.nn.ResNetBasicBlock(19, 19).cuda()
    # m(x_EEG)
    # model = MNetBackBorn(mngs.nn.MNet_config).cuda()
    # model(x_MEG)
    # Model
    BNet_config["n_chs_of_modalities"] = [160, 19]
    BNet_config["n_classes_of_modalities"] = [2, 4]    
    model = BNet(BNet_config, mngs.nn.MNet_config).cuda()

    # MEG
    y = model(x_MEG, 0)
    y = model(x_EEG, 1)    

    # # EEG
    # y = model(x_EEG)

    y.sum().backward()



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-05 02:05:43 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/general/_dotdict.py


# class ddict(dict):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)

#     def __getattr__(self, item):
#         return self[item]

#     def __setattr__(self, key, value):
#         self[key] = value

#     def __getstate__(self):
#         return self.__dict__

#     def __setstate__(self, state):
#         self.__dict__.update(state)
#         for key, value in state.items():
#             self[key] = value


# # class ddict(dict):
# #     def __init__(self, *args, **kwargs):
# #         super().__init__(*args, **kwargs)

# #     def __getattr__(self, item):
# #         return self[item]

# #     def __setattr__(self, key, value):
# #         self[key] = value

# import yaml


# def ddict_representer(dumper, data):
#     return dumper.represent_dict(data)


# def ddict_constructor(loader, node):
#     return ddict(loader.construct_pairs(node))


# yaml.add_representer(ddict, ddict_representer)
# yaml.add_constructor("tag:yaml.org,2002:map", ddict_constructor)


# class ddict(dict):
#     """Dot notation access to dictionary attributes"""

#     def __getattr__(self, item):
#         try:
#             return self[item]
#         except KeyError:
#             raise AttributeError(
#                 f"'{type(self).__name__}' object has no attribute '{item}'"
#             )

#     def __setattr__(self, key, value):
#         self[key] = value

#     def __delattr__(self, key):
#         try:
#             del self[key]
#         except KeyError:
#             raise AttributeError(
#                 f"'{type(self).__name__}' object has no attribute '{key}'"
#             )

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)


# class ddict(dict):
#     """dot.notation access to dictionary attributes"""

#     __getattr__ = dict.get
#     __setattr__ = dict.__setitem__
#     __delattr__ = dict.__delitem__



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-10 22:58:29 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_ChatGPT.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import sys

import google.generativeai as genai
import matplotlib.pyplot as plt
import mngs

from ._BaseGenAI import BaseGenAI

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


class Gemini(BaseGenAI):
    def __init__(
        self,
        system_setting="",
        api_key=os.getenv("GOOGLE_API_KEY"),
        model="gemini-1.5-pro-latest",
        stream=False,
        seed=None,
        n_keep=1,
        temperature=1.0,
    ):
        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key=api_key,
            stream=stream,
            seed=seed,
            n_keep=n_keep,
            temperature=temperature,
        )

    def _init_client(
        self,
    ):
        generation_config = genai.GenerationConfig(
            temperature=self.temperature
        )
        genai.configure(api_key=self.api_key)
        client = genai.GenerativeModel(
            self.model, generation_config=generation_config
        )
        chat_client = client.start_chat(history=self.chat_history)
        return chat_client

    def _api_call_static(
        self,
    ):
        prompt = self.chat_history[-1]["content"]
        response = self.client.send_message(prompt)
        out_text = response.text
        return out_text

    def _api_call_stream(
        self,
    ):
        prompt = self.chat_history[-1]["content"]
        responses = self.client.send_message(prompt, stream=True)
        for chunk in responses:
            yield chunk.text

    def _get_available_models(self):
        return [m.name for m in genai.list_models()]


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# Time-stamp: "2021-10-17 15:08:30 (ylab)"

from . import dsp, general, ml, plt, resource, stats

__copyright__ = "Copyright (C) 2021 Yusuke Watanabe"
__version__ = "0.0.29"
__license__ = "GPL3.0"
__author__ = "ywatanabe1989"
__author_email__ = "ywata1989@gmail.com"
__url__ = "https://github.com/ywatanabe1989/mngs"



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 19:11:48 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/pd/_misc.py


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import pandas as pd

# sys.path = ["."] + sys.path
# from scripts import utils, load


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def find_indi(df, conditions):
    # Generate a DataFrame where each column is a boolean Series representing one condition
    condition_df = pd.concat(
        [
            (df[k] == v) if not isinstance(v, list) else df[k].isin(v)
            for k, v in conditions.items()
        ],
        axis=1,
    )
    # Return a boolean Series where each value is True only if all conditions are met for that row
    return condition_df.all(axis=1)


def slice(df, conditions):
    _df = df.copy()
    # Use the find_indi function to get the indices of rows that meet all conditions
    indi = find_indi(df, conditions)
    # Return the filtered DataFrame
    return _df[indi]


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: 2024-05-14 00:03:52 (2)
# /ssh:ywatanabe@444:/home/ywatanabe/proj/mngs/src/mngs/plt/ax/_force_aspect.py

import matplotlib.pyplot as plt
import numpy as np


def force_aspect(ax, aspect=1):
    im = ax.get_images()

    extent = im[0].get_extent()

    ax.set_aspect(
        abs((extent[1] - extent[0]) / (extent[3] - extent[2])) / aspect
    )
    return ax


# Traceback (most recent call last):
#   File "/home/ywatanabe/proj/entrance/neurovista/./scripts/ml/clustering_vit.py", line 199, in <module>
#     main(args.model, args.clf_model)
#   File "/home/ywatanabe/proj/entrance/neurovista/./scripts/ml/clustering_vit.py", line 152, in main
#     fig, _legend_figs, _model = clustering_fn(
#   File "/home/ywatanabe/proj/mngs/src/mngs/ml/clustering/_pca.py", line 64, in pca
#     ax = mngs.plt.ax.force_aspect(ax)
#   File "/home/ywatanabe/proj/mngs/src/mngs/plt/ax/_force_aspect.py", line 13, in force_aspect
#     extent = im[0].get_extent()
# IndexError: list index out of range



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 19:10:36 (ywatanabe)"


def reload(module_or_func, verbose=False):
    """
    Reloads the given module or the module containing the given function.

    This function attempts to reload a module directly if a module is passed.
    If a function is passed, it tries to reload the module that contains the function.
    This can be useful during development when changes are made to modules and you want
    those changes to be reflected without restarting the Python interpreter.

    Arguments:
        module_or_func (module|function): The module or function to reload. If a function is
                                          provided, its containing module is reloaded.

    Returns:
        None

    Note:
        Reloading modules can have unexpected side effects, especially for modules that
        maintain state or for complex imports. Use with caution.
    """
    import importlib
    import sys

    if module_or_func in sys.modules:
        del sys.modules[module_or_func]
        importlib.reload(module_or_func)

    if hasattr(module_or_func, "__module__"):
        # If the object has a __module__ attribute, it's likely a function or class.
        # Attempt to reload its module.
        module_name = module_or_func.__module__
        if module_name not in sys.modules:
            print(
                f"Module {module_name} not found in sys.modules. Cannot reload."
            )
            return
    elif (
        hasattr(module_or_func, "__name__")
        and module_or_func.__name__ in sys.modules
    ):
        # Otherwise, assume it's a module and try to get its name directly.
        module_name = module_or_func.__name__
    else:
        print(
            f"Provided object is neither a recognized module nor a function/class with a __module__ attribute."
        )
        return

    try:
        # Attempt to reload the module by name.
        importlib.reload(sys.modules[module_name])

    except KeyError:
        # The module is not found in sys.modules, likely due to it not being imported.
        print(f"Module {module_name} not found in sys.modules. Cannot reload.")
    except Exception as e:
        # Catch any other exceptions and print an error message.
        print(f"Failed to reload module {module_name}. Error: {e}")



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-07 19:18:21 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/_subplots/FigWrapper.py

from mngs.gen import deprecated


class FigWrapper:
    """
    A wrapper class for a Matplotlib axis that collects plotting data.
    """

    def __init__(self, fig):
        """
        Initialize the AxisWrapper with a given axis and history reference.
        """
        self.fig = fig

    def __getattr__(self, attr):
        """
        Wrap the axis attribute access to collect plot calls or return the attribute directly.
        """
        original_attr = getattr(self.fig, attr)

        if callable(original_attr):

            def wrapper(
                *args, id=None, track=True, n_xticks=4, n_yticks=4, **kwargs
            ):
                result = original_attr(*args, **kwargs)

                return result

            return wrapper
        else:

            return original_attr

    ################################################################################
    # Original methods
    ################################################################################
    @deprecated("Use supxyt() instead.")
    def set_supxyt(self, *args, **kwargs):
        return self.supxyt(*args, **kwargs)

    def supxyt(self, xlabel=None, ylabel=None, title=None):
        """Sets xlabel, ylabel and title"""
        if xlabel is not None:
            self.fig.supxlabel(xlabel)
        if ylabel is not None:
            self.fig.supylabel(ylabel)
        if title is not None:
            self.fig.suptitle(title)
        return self.fig

    def tight_layout(self, rect=[0, 0.03, 1, 0.95]):
        self.fig.tight_layout(rect=rect)



#!/usr/bin/env python3

import fnmatch
import glob
import inspect
import os
import re
import subprocess
import warnings
from glob import glob

import mngs


################################################################################
## PATH
################################################################################
def get_this_fpath(when_ipython="/tmp/fake.py"):
    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = when_ipython  # "/tmp/fake.py"
    return __file__


def mk_spath(sfname, makedirs=False):

    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = f'/tmp/fake-{os.getenv("USER")}.py'

    ## spath
    fpath = __file__
    fdir, fname, _ = split_fpath(fpath)
    sdir = fdir + fname + "/"
    spath = sdir + sfname

    if makedirs:
        os.makedirs(mngs.path.split(spath)[0], exist_ok=True)

    return spath


def find_the_git_root_dir():
    import git

    repo = git.Repo(".", search_parent_directories=True)
    return repo.working_tree_dir


def split_fpath(fpath):
    """Split a file path to (1) the directory path, (2) the file name, and (3) the file extention
    Example:
        dirname, fname, ext = split_fpath('../data/01/day1/split_octave/2kHz_mat/tt8-2.mat')
        print(dirname) # '../data/01/day1/split_octave/2kHz_mat/'
        print(fname) # 'tt8-2'
        print(ext) # '.mat'
    """
    dirname = os.path.dirname(fpath) + "/"
    base = os.path.basename(fpath)
    fname, ext = os.path.splitext(base)
    return dirname, fname, ext


def touch(fpath):
    import pathlib

    return pathlib.Path(fpath).touch()


def find(rootdir, type="f", exp=["*"]):
    """
    Mimicks the Unix find command.

    Example:
        # rootdir =
        # type = 'f'  # 'f' for files, 'd' for directories, None for both
        # exp = '*.txt'  # Pattern to match, or None to match all
        find('/path/to/search', "f", "*.txt")
    """
    if isinstance(exp, str):
        exp = [exp]

    matches = []
    for _exp in exp:
        for root, dirs, files in os.walk(rootdir):
            # Depending on the type, choose the list to iterate over
            if type == "f":  # Files only
                names = files
            elif type == "d":  # Directories only
                names = dirs
            else:  # All entries
                names = files + dirs

            for name in names:
                # Construct the full path
                path = os.path.join(root, name)

                # If an _exp is provided, use fnmatch to filter names
                if _exp and not fnmatch.fnmatch(name, _exp):
                    continue

                # If type is set, ensure the type matches
                if type == "f" and not os.path.isfile(path):
                    continue
                if type == "d" and not os.path.isdir(path):
                    continue

                # Add the matching path to the results
                matches.append(path)

    return matches


def find_latest(dirname, fname, ext, version_prefix="_v"):
    version_pattern = re.compile(
        rf"({re.escape(fname)}{re.escape(version_prefix)})(\d+)({re.escape(ext)})$"
    )

    glob_pattern = os.path.join(dirname, f"{fname}{version_prefix}*{ext}")
    files = glob(glob_pattern)

    highest_version = 0
    latest_file = None

    for file in files:
        filename = os.path.basename(file)
        match = version_pattern.search(filename)
        if match:
            version_num = int(match.group(2))
            if version_num > highest_version:
                highest_version = version_num
                latest_file = file

    return latest_file


def increment_version(dirname, fname, ext, version_prefix="_v"):
    # Create a regex pattern to match the version number in the filename
    version_pattern = re.compile(
        rf"({re.escape(fname)}{re.escape(version_prefix)})(\d+)({re.escape(ext)})$"
    )

    # Construct the glob pattern to find all files that match the pattern
    glob_pattern = os.path.join(dirname, f"{fname}{version_prefix}*{ext}")

    # Use glob to find all files that match the pattern
    files = glob(glob_pattern)

    # Initialize the highest version number
    highest_version = 0
    base, suffix = None, None

    # Loop through the files to find the highest version number
    for file in files:
        filename = os.path.basename(file)
        match = version_pattern.search(filename)
        if match:
            base, version_str, suffix = match.groups()
            version_num = int(version_str)
            if version_num > highest_version:
                highest_version = version_num

    # If no versioned files were found, use the provided filename and extension
    if base is None or suffix is None:
        base = f"{fname}{version_prefix}"
        suffix = ext
        highest_version = 0  # No previous versions

    # Increment the highest version number
    next_version_number = highest_version + 1

    # Format the next version number with the same number of digits as the original
    next_version_str = f"{base}{next_version_number:03d}{suffix}"

    # Combine the directory and new filename to create the full path
    next_filepath = os.path.join(dirname, next_version_str)

    return next_filepath



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2021-12-12 14:50:37 (ywatanabe)"


from catboost import CatBoostClassifier
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.ensemble import AdaBoostClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import (
    LogisticRegression,
    PassiveAggressiveClassifier,
    Perceptron,
    RidgeClassifier,
    SGDClassifier,
)
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC, LinearSVC


class Classifiers(object):
    """Instanciates one of scikit-learn-like Clasifiers in the same manner.

    Example:
        clf_server = ClassifierServer(class_weight={0:1., 1:2.}, random_state=42)
        clf_str = "SVC"
        clf = clf_server(clf_str, scaler=StandardScaler())

    Note:
        clf_str is acceptable if it is in the list below.

        ['CatBoostClassifier',
         'Perceptron',
         'PassiveAggressiveClassifier',
         'LogisticRegression',
         'SGDClassifier',
         'RidgeClassifier',
         'QuadraticDiscriminantAnalysis',
         'GaussianProcessClassifier',
         'KNeighborsClassifier',
         'AdaBoostClassifier',
         'LinearSVC',
         'SVC']
    """

    def __init__(self, class_weight=None, random_state=42):
        self.class_weight = class_weight
        self.random_state = random_state

        self.clf_candi = {
            "CatBoostClassifier": CatBoostClassifier(
                class_weights=self.class_weight, verbose=False
            ),
            "Perceptron": Perceptron(
                penalty="l2", class_weight=self.class_weight, random_state=random_state
            ),
            "PassiveAggressiveClassifier": PassiveAggressiveClassifier(
                class_weight=self.class_weight, random_state=random_state
            ),
            "LogisticRegression": LogisticRegression(
                class_weight=self.class_weight, random_state=random_state
            ),
            "SGDClassifier": SGDClassifier(
                class_weight=self.class_weight, random_state=random_state
            ),
            "RidgeClassifier": RidgeClassifier(
                class_weight=self.class_weight, random_state=random_state
            ),
            "QuadraticDiscriminantAnalysis": QuadraticDiscriminantAnalysis(),
            "GaussianProcessClassifier": GaussianProcessClassifier(
                random_state=random_state
            ),
            "KNeighborsClassifier": KNeighborsClassifier(),
            "AdaBoostClassifier": AdaBoostClassifier(random_state=random_state),
            "LinearSVC": LinearSVC(
                class_weight=self.class_weight, random_state=random_state
            ),
            "SVC": SVC(class_weight=self.class_weight, random_state=random_state),
        }

    def __call__(self, clf_str, scaler=None):
        if scaler is not None:
            clf = make_pipeline(scaler, self.clf_candi[clf_str])  # fixme
        else:
            clf = self.clf_candi[clf_str]
        return clf

    @property
    def list(
        self,
    ):
        clf_list = list(self.clf_candi.keys())
        return clf_list


if __name__ == "__main__":
    clf_server = ClassifierServer()
    # l = clf_server.list
    clf = clf_server("SVC", scaler=StandardScaler())



#!/usr/bin/env python3

from ..ml.utils.grid_search import count_grids, yield_grids
from ._ci import ci
from ._close import close
from ._converters import (
    batch_fn,
    numpy_fn,
    squeeze_if,
    to_numpy,
    to_torch,
    torch_fn,
    unsqueeze_if,
)

# from ._ddict import ddict
from ._deprecated import deprecated
from ._dict_replace import dict_replace
from ._DimHandler import DimHandler
from ._embed import embed

# from ._find_indi import find_indi
from ._less import less
from ._mask_api import mask_api
from ._not_implemented import not_implemented

# Confirmed
from ._paste import paste
from ._src import src
from ._timeout import timeout
from ._title_case import title_case
from ._wrap import wrap

_ = None
from ..io.__init__ import *
from ._email import notify, send_gmail
from ._norm import to_z
from ._reproduce import fix_seeds, gen_ID, gen_timestamp
from ._shell import run_shellcommand, run_shellscript
from ._start import start
from ._tee import tee
from ._TimeStamper import TimeStamper
from ._title2path import title2path
from .latex import add_hat_in_the_latex_style, to_the_latex_style
from .misc import (
    _return_counting_process,
    color_text,
    connect_nums,
    connect_strs,
    copy_files,
    copy_the_file,
    ct,
    decapitalize,
    describe,
    find_closest,
    grep,
    is_defined_global,
    is_defined_local,
    is_later_or_equal,
    is_listed_X,
    is_nan,
    isclose,
    listed_dict,
    merge_dicts_wo_overlaps,
    mv_col,
    natglob,
    partial_at,
    pop_keys,
    print_block,
    readable_bytes,
    search,
    squeeze_spaces,
    suppress_output,
    symlink,
    to_even,
    to_odd,
    unique,
    uq,
    wait_key,
)
from .pandas import (
    col_to_last,
    col_to_top,
    force_dataframe,
    ignore_SettingWithCopyWarning,
    merge_columns,
)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-04 06:55:56 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/general/_ci.py


import numpy as np


def ci(xx, axis=None):
    indi = ~np.isnan(xx)
    return 1.96 * (xx[indi]).std(axis=axis) / np.sqrt(indi.sum())



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-05-16 13:19:50 (ywatanabe)"
# Author: Yusuke Watanabe (ywata1989@gmail.com)


"""
This script does XYZ.
"""


"""
Imports
"""
import math
import os
import sys
import time

import matplotlib.pyplot as plt
import mngs
import pandas as pd

"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def main(
    path="/tmp/mngs/processer_usages.csv",
    limit_min=3,
    interval_s=1,
    reset=True,
    verbose=True,
):
    # Parameters
    limit_s = limit_min * 60
    n_max = math.ceil(limit_s // interval_s)

    if reset and os.path.exists(path):
        mngs.sh(f"rm {path}")
        mngs.io.save(pd.DataFrame(), path, verbose=False)
        print(f"\n{path} was cleared.")

    for _ in range(n_max):
        add(path, verbose=verbose)
        time.sleep(interval_s)


def add(path, verbose=True):
    try:
        past = mngs.io.load(path)
    except Exception as e:
        print(e)
        past = pd.DataFrame()

    now = mngs.res.get_proc_usages()

    combined = pd.concat([past, now]).round(3)

    mngs.io.save(combined, path, verbose=verbose)

    if verbose:
        print(f"\n{combined}")


rec_procs = main

if __name__ == "__main__":
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-22 18:58:59"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

# Functions
# Imports
import inspect
import subprocess

import mngs

# # Config
# CONFIG = mngs.gen.load_configs()


def src(obj):
    """
    Returns the source code of a given object using `less`.
    Handles functions, classes, class instances, methods, and built-in functions.
    """
    # If obj is an instance of a class, get the class of the instance.
    if (
        not inspect.isclass(obj)
        and not inspect.isfunction(obj)
        and not inspect.ismethod(obj)
    ):
        obj = obj.__class__

    try:
        # Attempt to retrieve the source code
        source_code = inspect.getsource(obj)

        # Assuming mngs.gen.less is a placeholder for displaying text with `less`
        # This part of the code is commented out as it seems to be a placeholder
        # mngs.gen.less(source_code)

        # Open a subprocess to use `less` for displaying the source code
        process = subprocess.Popen(
            ["less"], stdin=subprocess.PIPE, encoding="utf8"
        )
        process.communicate(input=source_code)
        if process.returncode != 0:
            print(f"Process exited with return code {process.returncode}")
    except OSError as e:
        # Handle cases where the source code cannot be retrieved (e.g., built-in functions)
        print(f"Cannot retrieve source code: {e}")
    except TypeError as e:
        # Handle cases where the object type is not supported
        print(f"TypeError: {e}")
    except Exception as e:
        # Handle any other unexpected errors
        print(f"Error: {e}")


# def src(obj):
#     """
#     Returns the source code of a given object using `less`.
#     Handles functions, classes, class instances, and methods.
#     """
#     # If obj is an instance of a class, get the class of the instance.
#     if (
#         not inspect.isclass(obj)
#         and not inspect.isfunction(obj)
#         and not inspect.ismethod(obj)
#     ):
#         obj = obj.__class__

#     try:
#         # Attempt to retrieve the source code
#         source_code = inspect.getsource(obj)
#         mngs.gen.less(source_code)

#         # # Open a subprocess to use `less` for displaying the source code
#         # process = subprocess.Popen(
#         #     ["less"], stdin=subprocess.PIPE, encoding="utf8"
#         # )
#         # process.communicate(input=source_code)
#         if process.returncode != 0:
#             print(f"Process exited with return code {process.returncode}")
#     except TypeError as e:
#         # Handle cases where the object type is not supported
#         print(f"TypeError: {e}")
#     except Exception as e:
#         # Handle any other unexpected errors
#         print(f"Error: {e}")


# (YOUR AWESOME CODE)

if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/gen/_def.py
"""



#!/usr/bin/env python3
# Time-stamp: "2024-03-25 14:52:31 (ywatanabe)"

import os

import mngs
import numpy as np


class EarlyStopping:
    """
    Early stops the training if the validation score doesn't improve after a given patience period.

    """

    def __init__(
        self, patience=7, verbose=False, delta=1e-5, direction="minimize"
    ):
        """
        Args:
            patience (int): How long to wait after last time validation score improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation score improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
        """
        self.patience = patience
        self.verbose = verbose
        self.direction = direction

        self.delta = delta

        # default
        self.counter = 0
        self.best_score = np.Inf if direction == "minimize" else -np.Inf
        self.best_i_global = None
        self.models_spaths_dict = {}

    def is_best(self, val_score):
        is_smaller = val_score < self.best_score - abs(self.delta)
        is_larger = self.best_score + abs(self.delta) < val_score
        return is_smaller if self.direction == "minimize" else is_larger

    def __call__(self, current_score, models_spaths_dict, i_global):
        # The 1st call
        if self.best_score is None:
            self.save(current_score, models_spaths_dict, i_global)
            return False

        # After the 2nd call
        if self.is_best(current_score):
            self.save(current_score, models_spaths_dict, i_global)
            self.counter = 0
            return False

        else:
            self.counter += 1
            if self.verbose:
                print(
                    f"\nEarlyStopping counter: {self.counter} out of {self.patience}\n"
                )
            if self.counter >= self.patience:
                if self.verbose:
                    mngs.gen.print_block("Early-stopped.", c="yellow")
                return True

    def save(self, current_score, models_spaths_dict, i_global):
        """Saves model when validation score decrease."""

        if self.verbose:
            print(
                f"\nUpdate the best score: ({self.best_score:.6f} --> {current_score:.6f})"
            )

        self.best_score = current_score
        self.best_i_global = i_global

        for model, spath in models_spaths_dict.items():
            mngs.io.save(model.state_dict(), spath)

        self.models_spaths_dict = models_spaths_dict


if __name__ == "__main__":
    pass
    # # starts the current fold's loop
    # i_global = 0
    # lc_logger = mngs.ml.LearningCurveLogger()
    # early_stopping = utils.EarlyStopping(patience=50, verbose=True)
    # for i_epoch, epoch in enumerate(tqdm(range(merged_conf["MAX_EPOCHS"]))):

    #     dlf.fill(i_fold, reset_fill_counter=False)

    #     step_str = "Validation"
    #     for i_batch, batch in enumerate(dlf.dl_val):
    #         _, loss_diag_val = utils.base_step(
    #             step_str,
    #             model,
    #             mtl,
    #             batch,
    #             device,
    #             i_fold,
    #             i_epoch,
    #             i_batch,
    #             i_global,
    #             lc_logger,
    #             no_mtl=args.no_mtl,
    #             print_batch_interval=False,
    #         )
    #     lc_logger.print(step_str)

    #     step_str = "Training"
    #     for i_batch, batch in enumerate(dlf.dl_tra):
    #         optimizer.zero_grad()
    #         loss, _ = utils.base_step(
    #             step_str,
    #             model,
    #             mtl,
    #             batch,
    #             device,
    #             i_fold,
    #             i_epoch,
    #             i_batch,
    #             i_global,
    #             lc_logger,
    #             no_mtl=args.no_mtl,
    #             print_batch_interval=False,
    #         )
    #         loss.backward()
    #         optimizer.step()
    #         i_global += 1
    #     lc_logger.print(step_str)

    #     bACC_val = np.array(lc_logger.logged_dict["Validation"]["bACC_diag_plot"])[
    #         np.array(lc_logger.logged_dict["Validation"]["i_epoch"]) == i_epoch
    #     ].mean()

    #     model_spath = (
    #         merged_conf["sdir"]
    #         + f"checkpoints/model_fold#{i_fold}_epoch#{i_epoch:03d}.pth"
    #     )
    #     mtl_spath = model_spath.replace("model_fold", "mtl_fold")
    #     models_spaths_dict = {model_spath: model, mtl_spath: mtl}

    #     early_stopping(loss_diag_val, models_spaths_dict, i_epoch, i_global)
    #     # early_stopping(-bACC_val, models_spaths_dict, i_epoch, i_global)

    #     if early_stopping.early_stop:
    #         print("Early stopping")
    #         break



#!/usr/bin/env python3

import time

import pandas as pd


class TimeStamper:
    """
    A class for generating timestamps with optional comments, tracking both the time since object creation and since the last call.
    """

    def __init__(self, is_simple=True):
        self.id = -1
        self.start_time = time.time()
        self._is_simple = is_simple
        self._prev = self.start_time
        self._df_record = pd.DataFrame(
            columns=[
                "timestamp",
                "elapsed_since_start",
                "elapsed_since_prev",
                "comment",
                "formatted_text",
            ]
        )

    def __call__(self, comment="", verbose=False):
        now = time.time()
        from_start = now - self.start_time
        from_prev = now - self._prev
        formatted_from_start = time.strftime(
            "%H:%M:%S", time.gmtime(from_start)
        )
        formatted_from_prev = time.strftime("%H:%M:%S", time.gmtime(from_prev))
        self.id += 1
        self._prev = now
        text = (
            f"ID:{self.id} | {formatted_from_start} {comment} | "
            if self._is_simple
            else f"Time (id:{self.id}): total {formatted_from_start}, prev {formatted_from_prev} [hh:mm:ss]: {comment}\n"
        )

        # Update DataFrame directly
        self._df_record.loc[self.id] = [
            now,
            from_start,
            from_prev,
            comment,
            text,
        ]

        if verbose:
            print(text)
        return text

    @property
    def record(self):
        return self._df_record[
            [
                "timestamp",
                "elapsed_since_start",
                "elapsed_since_prev",
                "comment",
            ]
        ]

    def delta(self, id1, id2):
        """
        Calculate the difference in seconds between two timestamps identified by their IDs.

        Parameters:
            id1 (int): The ID of the first timestamp.
            id2 (int): The ID of the second timestamp.

        Returns:
            float: The difference in seconds between the two timestamps.

        Raises:
            ValueError: If either id1 or id2 is not in the DataFrame index.
        """
        # Adjust for negative indices, similar to negative list indexing in Python
        if id1 < 0:
            id1 = len(self._df_record) + id1
        if id2 < 0:
            id2 = len(self._df_record) + id2

        # Check if both IDs exist in the DataFrame
        if (
            id1 not in self._df_record.index
            or id2 not in self._df_record.index
        ):
            raise ValueError(
                "One or both of the IDs do not exist in the record."
            )

        # Compute the difference in timestamps
        time_diff = (
            self._df_record.loc[id1, "timestamp"]
            - self._df_record.loc[id2, "timestamp"]
        )
        return time_diff


# class TimeStamper:
#     """
#     A class for generating timestamps with optional comments, tracking both the time since object creation and since the last call.

#     Attributes:
#         id (int): An identifier for each timestamp, starting from 0.
#         start_time (float): The time when the TimeStamper object was created.

#     Arguments:
#         is_simple (bool): Determines the output format. Defaults to True for a simpler format.

#     Returns:
#         str: The generated timestamp string.

#     Example:
#         >>> ts = TimeStamper(is_simple=True)
#         >>> print(ts("Starting process"))
#         ID:0 | 00:00:00 Starting process |
#         >>> time.sleep(1)
#         >>> print(ts("One second later"))
#         ID:1 | 00:00:01 One second later |
#     """

#     def __init__(self, is_simple=True):
#         self.id = -1
#         self.start_time = time.time()
#         self._is_simple = is_simple
#         self._prev = self.start_time
#         self._record = []

#     def __call__(self, comment=""):
#         # Calculation
#         now = time.time()
#         from_start = now - self.start_time
#         from_prev = now - self._prev

#         # Format time strings for display
#         formatted_from_start = time.strftime(
#             "%H:%M:%S", time.gmtime(from_start)
#         )
#         formatted_from_prev = time.strftime("%H:%M:%S", time.gmtime(from_prev))

#         # Increment ID
#         self.id += 1
#         self._prev = now

#         # Text construction
#         if self._is_simple:
#             text = f"ID:{self.id} | {formatted_from_start} {comment} | "
#         else:
#             text = f"Time (id:{self.id}): total {formatted_from_start}, prev {formatted_from_prev} [hh:mm:ss]: {comment}\n"

#         # Update record with structured data
#         self._record.append(
#             {
#                 "id": self.id,
#                 "timestamp": now,
#                 "elapsed_since_start": from_start,
#                 "elapsed_since_prev": from_prev,
#                 "comment": comment,
#                 "formatted_text": text,
#             }
#         )

#         print(text)
#         return text

#     @property
#     def record(
#         self,
#     ):
#         cols = [
#             "timestamp",
#             "elapsed_since_start",
#             "elapsed_since_prev",
#             "comment",
#         ]
#         return pd.DataFrame(self._record).set_index("id")[cols]


# class TimeStamper:
#     """
#     A class for generating timestamps with optional comments, tracking both the time since object creation and since the last call.

#     Attributes:
#         id (int): An identifier for each timestamp, starting from 0.
#         start_time (float): The time when the TimeStamper object was created.

#     Arguments:
#         is_simple (bool): Determines the output format. Defaults to True for a simpler format.

#     Returns:
#         str: The generated timestamp string.

#     Example:
#         >>> ts = TimeStamper(is_simple=True)
#         >>> ts("Starting process")
#         ID:0 | 00:00:00 Starting process |
#         >>> time.sleep(1)
#         >>> ts("One second later")
#         ID:1 | 00:00:01 One second later |
#     """

#     def __init__(self, is_simple=True):
#         self.id = -1
#         self.start_time = time.time()
#         self._is_simple = is_simple
#         self._time = time
#         self._prev = self.start_time
#         self.record = []

#     def __call__(self, comment=""):
#         # Calculation
#         now = self._time.time()
#         from_start = now - self.start_time
#         from_prev = now - self._prev

#         # Saves as attributes
#         self._from_start_hhmmss = self._time.strftime(
#             "%H:%M:%S", self._time.gmtime(from_start)
#         )
#         self._from_prev_hhmmss = self._time.strftime(
#             "%H:%M:%S", self._time.gmtime(from_prev)
#         )

#         # Incrementation
#         self.id += 1
#         self.prev = now

#         # Text construction
#         if self._is_simple:
#             self.text = (
#                 f"ID:{self.id} | {self._from_start_hhmmss} {comment} | "
#             )
#         else:
#             self.text = f"Time (id:{self.id}): total {self._from_start_hhmmss}, prev {self._from_prev_hhmmss} [hh:mm:ss]: {comment}\n"

#         print(self.text)
#         self.record.append(self.text)
#         return self.text


if __name__ == "__main__":
    ts = TimeStamper(is_simple=True)
    ts("Starting process")
    time.sleep(1)
    ts("One second later")
    time.sleep(2)
    ts("Two seconds later")



import mngs
import numpy as np
import matplotlib
# matplotlib.use("TkAgg")
import matplotlib.pyplot as plt


def mk_colorbar(start="white", end="blue"):
    xx = np.linspace(0, 1, 256)

    start = np.array(mngs.plt.colors.RGB_d[start])
    end = np.array(mngs.plt.colors.RGB_d[end])
    colors = (end-start)[:, np.newaxis]*xx

    colors -= colors.min()
    colors /= colors.max()

    fig, ax = plt.subplots()
    [ax.axvline(_xx, color=colors[:,i_xx]) for i_xx, _xx in enumerate(xx)]
    ax.xaxis.set_ticks_position("none")
    ax.yaxis.set_ticks_position("none")
    ax.set_aspect(0.2)
    return fig





#!/usr/bin/env python3

import mngs


################################################################################
## Reproducibility
################################################################################
def fix_seeds(
    os=None, random=None, np=None, torch=None, tf=None, seed=42, verbose=True
):
    os_str = "os" if os is not None else ""
    random_str = "random" if random is not None else ""
    np_str = "np" if np is not None else ""
    torch_str = "torch" if torch is not None else ""
    tf_str = "tf" if tf is not None else ""

    # https://github.com/lucidrains/vit-pytorch/blob/main/examples/cats_and_dogs.ipynb
    if os is not None:
        import os

        os.environ["PYTHONHASHSEED"] = str(seed)

    if random is not None:
        random.seed(seed)

    if np is not None:
        np.random.seed(seed)

    if torch is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        # torch.use_deterministic_algorithms(True)

    if tf is not None:
        tf.random.set_seed(seed)

    if verbose:
        print(f"\n{'-'*40}\n")
        print(
            f"Random seeds of the following packages have been fixed as {seed}"
        )
        print(os_str, random_str, np_str, torch_str, tf_str)
        print(f"\n{'-'*40}\n")


def gen_ID(time_format="%YY-%mM-%dD-%Hh%Mm%Ss", N=8):
    import random
    import string
    from datetime import datetime

    now = datetime.now()
    # now_str = now.strftime("%Y-%m-%d-%H-%M")
    now_str = now.strftime(time_format)

    # today_str = now.strftime("%Y-%m%d")
    randlst = [
        random.choice(string.ascii_letters + string.digits) for i in range(N)
    ]
    rand_str = "".join(randlst)
    return now_str + "_" + rand_str


def gen_timestamp():
    from datetime import datetime

    now = datetime.now()
    now_str = now.strftime("%Y-%m%d-%H%M")
    return now_str
